"""
Implementation of Prefix Delegation options as specified in :rfc:`3633`.
"""

from ipaddress import IPv6Address, IPv6Network
from struct import unpack_from, pack

from dhcpkit.ipv6.messages import SolicitMessage, AdvertiseMessage, RequestMessage, RenewMessage, \
    RebindMessage, ReleaseMessage, ReplyMessage
from dhcpkit.ipv6.options import Option, StatusCodeOption

OPTION_IA_PD = 25
OPTION_IAPREFIX = 26

STATUS_NOPREFIXAVAIL = 6


class IAPDOption(Option):
    """
    :rfc:`3633#section-9`

    The IA_PD option is used to carry a prefix delegation identity
    association, the parameters associated with the IA_PD and the
    prefixes associated with it.

    The format of the IA_PD option is::

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |         OPTION_IA_PD          |         option-length         |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                         IAID (4 octets)                       |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                              T1                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                              T2                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      .                                                               .
      .                          IA_PD-options                        .
      .                                                               .
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    option-code
        OPTION_IA_PD (25).

    option-length
        12 + length of IA_PD-options field.

    IAID
        The unique identifier for this IA_PD; the IAID must be unique among the identifiers for all of this requesting
        router's IA_PDs.

    T1
        The time at which the requesting router should contact the delegating router from which the prefixes in the
        IA_PD were obtained to extend the lifetimes of the prefixes delegated to the IA_PD; T1 is a time duration
        relative to the current time expressed in units of seconds.

    T2
        The time at which the requesting router should contact any available delegating router to extend the lifetimes
        of the prefixes assigned to the IA_PD; T2 is a time duration relative to the current time expressed in units of
        seconds.

    IA_PD-options
        Options associated with this IA_PD.

    The IA_PD-options field encapsulates those options that are specific
    to this IA_PD.  For example, all of the IA_PD Prefix Options carrying
    the prefixes associated with this IA_PD are in the IA_PD-options
    field.

    An IA_PD option may only appear in the options area of a DHCP
    message.  A DHCP message may contain multiple IA_PD options.

    The status of any operations involving this IA_PD is indicated in a
    Status Code option in the IA_PD-options field.

    Note that an IA_PD has no explicit "lifetime" or "lease length" of
    its own.  When the valid lifetimes of all of the prefixes in a IA_PD
    have expired, the IA_PD can be considered as having expired.  T1 and
    T2 are included to give delegating routers explicit control over when
    a requesting router should contact the delegating router about a
    specific IA_PD.

    In a message sent by a requesting router to a delegating router,
    values in the T1 and T2 fields indicate the requesting router's
    preference for those parameters.  The requesting router sets T1 and
    T2 to zero if it has no preference for those values.  In a message
    sent by a delegating router to a requesting router, the requesting
    router MUST use the values in the T1 and T2 fields for the T1 and T2
    parameters.  The values in the T1 and T2 fields are the number of
    seconds until T1 and T2.

    The delegating router selects the T1 and T2 times to allow the
    requesting router to extend the lifetimes of any prefixes in the
    IA_PD before the lifetimes expire, even if the delegating router is
    unavailable for some short period of time.  Recommended values for T1
    and T2 are .5 and .8 times the shortest preferred lifetime of the
    prefixes in the IA_PD that the delegating router is willing to
    extend, respectively.  If the time at which the prefixes in an IA_PD
    are to be renewed is to be left to the discretion of the requesting
    router, the delegating router sets T1 and T2 to 0.

    If a delegating router receives an IA_PD with T1 greater than T2, and
    both T1 and T2 are greater than 0, the delegating router ignores the
    invalid values of T1 and T2 and processes the IA_PD as though the
    delegating router had set T1 and T2 to 0.

    If a requesting router receives an IA_PD with T1 greater than T2, and
    both T1 and T2 are greater than 0, the client discards the IA_PD
    option and processes the remainder of the message as though the
    delegating router had not included the IA_PD option.

    :type iaid: bytes
    :type t1: int
    :type t2: int
    :type options: list[Option]
    """

    option_type = OPTION_IA_PD

    def __init__(self, iaid: bytes = b'\x00\x00\x00\x00', t1: int = 0, t2: int = 0, options: [Option] = None):
        self.iaid = iaid
        """The unique identifier for this IA_PD"""

        self.t1 = t1
        """The time at which the client contacts the server to renew its prefixes"""

        self.t2 = t2
        """The time at which the client contacts any available server to rebind its prefixes"""

        self.options = options or []
        """The list of options contained in this IAPDOption"""

    def validate(self):
        """
        Validate that the contents of this object conform to protocol specs.
        """
        if not isinstance(self.iaid, bytes) or len(self.iaid) != 4:
            raise ValueError("IAID must be four bytes")

        if not isinstance(self.t1, int) or not (0 <= self.t1 < 2 ** 32):
            raise ValueError("T1 must be an unsigned 32 bit integer")

        if not isinstance(self.t2, int) or not (0 <= self.t2 < 2 ** 32):
            raise ValueError("T2 must be an unsigned 32 bit integer")

        # Check if all options are allowed
        self.validate_contains(self.options)
        for option in self.options:
            option.validate()

    def load_from(self, buffer: bytes, offset: int = 0, length: int = None) -> int:
        """
        Load the internal state of this object from the given buffer. The buffer may contain more data after the
        structured element is parsed. This data is ignored.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """
        my_offset, option_len = self.parse_option_header(buffer, offset, length)
        header_offset = my_offset

        self.iaid = buffer[offset + my_offset:offset + my_offset + 4]
        my_offset += 4

        self.t1, self.t2 = unpack_from('!II', buffer, offset + my_offset)
        my_offset += 8

        # Parse the options
        self.options = []
        max_offset = option_len + header_offset  # The option_len field counts bytes *after* the header fields
        while max_offset > my_offset:
            used_buffer, option = Option.parse(buffer, offset=offset + my_offset)
            self.options.append(option)
            my_offset += used_buffer

        if my_offset != max_offset:
            raise ValueError('Option length does not match the combined length of the parsed options')

        self.validate()

        return my_offset

    def save(self) -> bytes:
        """
        Save the internal state of this object as a buffer.

        :return: The buffer with the data from this element
        """
        self.validate()
        options_buffer = bytearray()
        for option in self.options:
            options_buffer.extend(option.save())

        buffer = bytearray()
        buffer.extend(pack('!HH4sII', self.option_type, len(options_buffer) + 12, self.iaid, self.t1, self.t2))
        buffer.extend(options_buffer)
        return buffer

    def get_options_of_type(self, klass: type) -> list:
        """
        Get all options that are subclasses of the given class.

        :param klass: The class to look for
        :returns: The list of options

        :type klass: T
        :rtype: list[T()]
        """
        return [option for option in self.options if isinstance(option, klass)]

    def get_option_of_type(self, klass: type) -> object or None:
        """
        Get the first option that is a subclass of the given class.

        :param klass: The class to look for
        :returns: The option or None

        :type klass: T
        :rtype: T()
        """
        for option in self.options:
            if isinstance(option, klass):
                return option

    def get_prefixes(self) -> [IPv6Network]:
        """
        Get all prefixes from IAPrefixOptions

        :returns: list if prefixes
        """
        return [suboption.prefix for suboption in self.get_options_of_type(IAPrefixOption)]


class IAPrefixOption(Option):
    """
    :rfc:`3633#section-10`

    The IA_PD Prefix option is used to specify IPv6 address prefixes
    associated with an IA_PD.  The IA_PD Prefix option must be
    encapsulated in the IA_PD-options field of an IA_PD option.

    The format of the IA_PD Prefix option is::

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |        OPTION_IAPREFIX        |         option-length         |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                      preferred-lifetime                       |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                        valid-lifetime                         |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      | prefix-length |                                               |
      +-+-+-+-+-+-+-+-+          IPv6 prefix                          |
      |                           (16 octets)                         |
      |                                                               |
      |                                                               |
      |                                                               |
      |               +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |               |                                               .
      +-+-+-+-+-+-+-+-+                                               .
      .                       IAprefix-options                        .
      .                                                               .
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    option-code
        OPTION_IAPREFIX (26).

    option-length
        25 + length of IAprefix-options field.

    preferred-lifetime
     The recommended preferred lifetime for the IPv6 prefix in the option, expressed in units of seconds. A value of
     0xFFFFFFFF represents infinity.

    valid-lifetime
       The valid lifetime for the IPv6 prefix in the option, expressed in units of seconds.  A value of 0xFFFFFFFF
       represents infinity.

    prefix-length
        Length for this prefix in bits.

    IPv6-prefix
        An IPv6 prefix.

    IAprefix-options
        Options associated with this prefix.

    In a message sent by a requesting router to a delegating router, the
    values in the fields can be used to indicate the requesting router's
    preference for those values.  The requesting router may send a value
    of zero to indicate no preference.  A requesting router may set the
    IPv6 prefix field to zero and a given value in the prefix-length
    field to indicate a preference for the size of the prefix to be
    delegated.

    In a message sent by a delegating router the preferred and valid
    lifetimes should be set to the values of AdvPreferredLifetime and
    AdvValidLifetime as specified in section 6.2.1, "Router Configuration
    Variables" of :rfc:`2461` [4], unless administratively configured.

    A requesting router discards any prefixes for which the preferred
    lifetime is greater than the valid lifetime.  A delegating router
    ignores the lifetimes set by the requesting router if the preferred
    lifetime is greater than the valid lifetime and ignores the values
    for T1 and T2 set by the requesting router if those values are
    greater than the preferred lifetime.

    The values in the preferred and valid lifetimes are the number of
    seconds remaining for each lifetime.

    An IA_PD Prefix option may appear only in an IA_PD option.  More than
    one IA_PD Prefix Option can appear in a single IA_PD option.

    The status of any operations involving this IA_PD Prefix option is
    indicated in a Status Code option in the IAprefix-options field.

    :type preferred_lifetime: int
    :type valid_lifetime: int
    :type prefix: IPv6Network
    :type options: list[Option]
    """

    option_type = OPTION_IAPREFIX

    def __init__(self, prefix: IPv6Network = None, preferred_lifetime: int = 0, valid_lifetime: int = 0,
                 options: [Option] = None):
        self.prefix = prefix
        """The IPv6 prefix"""

        self.preferred_lifetime = preferred_lifetime
        """The preferred lifetime of this IPv6 prefix"""

        self.valid_lifetime = valid_lifetime
        """The valid lifetime of this IPv6 prefix"""

        self.options = options or []
        """The list of options related to this IAPrefixOption"""

    def validate(self):
        """
        Validate that the contents of this object conform to protocol specs.
        """
        if not isinstance(self.preferred_lifetime, int) or not (0 <= self.preferred_lifetime < 2 ** 32):
            raise ValueError("Preferred lifetime must be an unsigned 32 bit integer")

        if not isinstance(self.valid_lifetime, int) or not (0 <= self.valid_lifetime < 2 ** 32):
            raise ValueError("Valid lifetime must be an unsigned 32 bit integer")

        if not isinstance(self.prefix, IPv6Network) or self.prefix.is_link_local or self.prefix.is_loopback \
                or self.prefix.is_multicast:
            raise ValueError("Prefix must be a routable IPv6 prefix")

        # Check if all options are allowed
        self.validate_contains(self.options)
        for option in self.options:
            option.validate()

    def load_from(self, buffer: bytes, offset: int = 0, length: int = None) -> int:
        """
        Load the internal state of this object from the given buffer. The buffer may contain more data after the
        structured element is parsed. This data is ignored.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """
        my_offset, option_len = self.parse_option_header(buffer, offset, length)
        header_offset = my_offset

        self.preferred_lifetime, self.valid_lifetime = unpack_from('!II', buffer, offset=offset + my_offset)
        my_offset += 8

        prefix_length = buffer[offset + my_offset]
        my_offset += 1

        address = IPv6Address(buffer[offset + my_offset:offset + my_offset + 16])
        my_offset += 16

        # Combine address and prefix length into prefix
        self.prefix = IPv6Network('{!s}/{:d}'.format(address, prefix_length))

        # Parse the options
        self.options = []
        max_offset = option_len + header_offset  # The option_len field counts bytes *after* the header fields
        while max_offset > my_offset:
            used_buffer, option = Option.parse(buffer, offset=offset + my_offset)
            self.options.append(option)
            my_offset += used_buffer

        if my_offset != max_offset:
            raise ValueError('Option length does not match the combined length of the parsed options')

        self.validate()

        return my_offset

    def save(self) -> bytes:
        """
        Save the internal state of this object as a buffer.

        :return: The buffer with the data from this element
        """
        self.validate()

        options_buffer = bytearray()
        for option in self.options:
            options_buffer.extend(option.save())

        buffer = bytearray()
        buffer.extend(pack('!HHIIB', self.option_type, len(options_buffer) + 25,
                           self.preferred_lifetime, self.valid_lifetime, self.prefix.prefixlen))
        buffer.extend(self.prefix.network_address.packed)
        buffer.extend(options_buffer)
        return buffer


# Register where these options may occur
SolicitMessage.add_may_contain(IAPDOption)
AdvertiseMessage.add_may_contain(IAPDOption)
RequestMessage.add_may_contain(IAPDOption)
RenewMessage.add_may_contain(IAPDOption)
RebindMessage.add_may_contain(IAPDOption)
ReleaseMessage.add_may_contain(IAPDOption)
ReplyMessage.add_may_contain(IAPDOption)

IAPDOption.add_may_contain(IAPrefixOption)
IAPDOption.add_may_contain(StatusCodeOption, 0, 1)

IAPrefixOption.add_may_contain(StatusCodeOption, 0, 1)



"""
License for this file:

Copyright (c) 2011 Mateusz Kobos

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from contextlib import contextmanager
import threading

__author__ = "Mateusz Kobos"


class RWLock:
    """
    Synchronization object used in a solution of so-called second
    readers-writers problem. In this problem, many readers can simultaneously
    access a share, and a writer has an exclusive access to this share.

    Additionally, the following constraints should be met:

    1) no reader should be kept waiting if the share is currently opened for
       reading unless a writer is also waiting for the share,
    2) no writer should be kept waiting for the share longer than absolutely
       necessary.

    The implementation is based on [1]_
    with a modification -- adding an additional lock (:attr:`__readers_queue`)
    -- in accordance with [2]_.

    .. rubric:: Footnotes

    .. [1] A.B. Downey: "The little book of semaphores", Version 2.1.5, 2008,
           See sections 4.2.2, 4.2.6, 4.2.7
    .. [2] P.J. Courtois, F. Heymans, D.L. Parnas:
           "Concurrent Control with 'Readers' and 'Writers'",
           Communications of the ACM, 1971 (via [3]_)
    .. [3] http://en.wikipedia.org/wiki/Readers-writers_problem

    :type __read_switch: _LightSwitch
    :type __write_switch: _LightSwitch
    :type __no_readers: threading.Lock
    :type __no_writers: threading.Lock
    :type __readers_queue: threading.Lock
    """

    def __init__(self):
        self.__read_switch = _LightSwitch()
        self.__write_switch = _LightSwitch()
        self.__no_readers = threading.Lock()
        self.__no_writers = threading.Lock()
        self.__readers_queue = threading.Lock()
        """A lock giving an even higher priority to the writer in certain
        cases (see [2] for a discussion)"""

    @property
    def readers(self) -> int:
        """
        The number of active readers holding this lock

        :return: The number of readers
        """
        return self.__read_switch.counter

    @property
    def writers(self):
        """
        The number of writers holding this lock. May be more than one if there is a queue.

        :return: The number of readers
        """
        return self.__write_switch.counter

    @property
    def blocked_for_readers(self):
        """
        Whether this lock is blocked for readers

        :return: Whether a reader would block when trying to acquire this lock
        """
        have_locked = self.__no_readers.acquire(blocking=False)
        if have_locked:
            # It wasn't locked and we locked it. Let go immediately
            self.__no_readers.release()
        return not have_locked

    @property
    def blocked_for_writers(self):
        """
        Whether this lock is blocked for writers

        :return: Whether a writer would block when trying to acquire this lock
        """
        have_locked = self.__no_writers.acquire(blocking=False)
        if have_locked:
            # It wasn't locked and we locked it. Let go immediately
            self.__no_writers.release()
        return not have_locked

    def reader_acquire(self):
        """
        Acquire the lock as a reader.
        """
        self.__readers_queue.acquire()
        self.__no_readers.acquire()
        self.__read_switch.acquire(self.__no_writers)
        self.__no_readers.release()
        self.__readers_queue.release()

    def reader_release(self):
        """
        Release the reader lock. Must have been acquired first.
        """
        self.__read_switch.release(self.__no_writers)

    def writer_acquire(self):
        """
        Acquire the lock as a writer.
        """
        self.__write_switch.acquire(self.__no_readers)
        self.__no_writers.acquire()

    def writer_release(self):
        """
        Release the writer lock. Must have been acquired first.
        """
        self.__no_writers.release()
        self.__write_switch.release(self.__no_readers)

    @contextmanager
    def read_lock(self):
        """
        A context-manager that acquires a reader lock while in-context.
        """
        self.reader_acquire()
        yield
        self.reader_release()

    @contextmanager
    def write_lock(self):
        """
        A context-manager that acquires a writer lock while in-context.
        """
        self.writer_acquire()
        yield
        self.writer_release()


class _LightSwitch:
    """
    An auxiliary "light switch"-like object. The first thread turns on the
    "switch", the last one turns it off (see [1, sec. 4.2.2] for details).

    :type __counter: int
    :type __mutex: threading.Lock
    """

    def __init__(self):
        self.__counter = 0
        self.__mutex = threading.Lock()

    @property
    def counter(self):
        """
        The number of 'people in the room'. First person entering will turn on the 'switch' (Lock) and the last person
        leaving will turn of the switch/Lock.

        :return: The number of users of this light switch
        """
        return self.__counter

    def acquire(self, lock):
        """
        'Enter the room' equivalent. If this is the first person entering then 'switch the light on' (Lock the lock)

        :param lock: The lock to lock
        """
        self.__mutex.acquire()
        self.__counter += 1
        if self.__counter == 1:
            lock.acquire()
        self.__mutex.release()

    def release(self, lock):
        """
        'Leave the room' equivalent. If this is the last person leaving then 'switch the light off' (Unlock the lock)

        :param lock: The lock to lock
        """
        self.__mutex.acquire()
        self.__counter -= 1
        if self.__counter == 0:
            lock.release()
        self.__mutex.release()



"""
Classes that handle the processing of an option
"""
import abc
import configparser
import logging

from dhcpkit.ipv6.messages import RelayReplyMessage, RelayForwardMessage
from dhcpkit.ipv6.options import OptionRequestOption, Option
from dhcpkit.ipv6.transaction_bundle import TransactionBundle

logger = logging.getLogger(__name__)


class OptionHandler(metaclass=abc.ABCMeta):
    """
    Base class for option handlers
    """

    @classmethod
    def from_config(cls, section: configparser.SectionProxy, option_handler_id: str = None) -> object:
        """
        Create a handler of this class based on the configuration in the config section. No default implementation
        is provided. Subclasses should implement their own if they want to be loaded from a configuration file.

        :param section: The configuration section
        :param option_handler_id: Optional extra identifier for an option handler
        :return: A handler object
        :rtype: OptionHandler
        """
        raise configparser.Error("{} does not support loading from configuration".format(cls.__name__))

    # noinspection PyMethodMayBeStatic
    def pre(self, bundle: TransactionBundle):
        """
        Pre-process the data in the bundle. Subclasses can update bundle state here or abort processing of the request
        by raising a CannotRespondError.

        :param bundle: The transaction bundle
        """

    # noinspection PyMethodMayBeStatic
    @abc.abstractmethod
    def handle(self, bundle: TransactionBundle):
        """
        Handle the data in the bundle. Should do their main work here.

        :param bundle: The transaction bundle
        """

    # noinspection PyMethodMayBeStatic
    def post(self, bundle: TransactionBundle):
        """
        Post-process the data in the bundle. Subclasses can e.g. clean up state. Subclasses assigning addresses should
        check whether the bundle.response is a AdvertiseMessage or a ReplyMessage. The class can change between handle()
        and post() when the server is using rapid-commit.

        :param bundle: The transaction bundle
        """


class RelayOptionHandler(OptionHandler):
    """
    A base class for option handlers that work on option in the relay messages chain.
    """

    def handle(self, bundle: TransactionBundle):
        """
        Handle the data in the bundle by checking the relay chain and calling :meth:`handle_relay` for each relay
        message.

        :param bundle: The transaction bundle
        """
        # We need the outgoing chain to be present
        if bundle.outgoing_relay_messages is None:
            logger.error("Cannot process relay chains: outgoing chain not set")
            return

        # Don't try to match between chains of different size
        if len(bundle.incoming_relay_messages) != len(bundle.outgoing_relay_messages):
            logger.error("Cannot process relay chains: chain have different length")
            return

        # Process the relay messages one by one
        for relay_message_in, relay_message_out in zip(bundle.incoming_relay_messages, bundle.outgoing_relay_messages):
            self.handle_relay(bundle, relay_message_in, relay_message_out)

    @abc.abstractmethod
    def handle_relay(self, bundle: TransactionBundle,
                     relay_message_in: RelayForwardMessage, relay_message_out: RelayReplyMessage):
        """
        Handle the options for each relay message pair.

        :param bundle: The transaction bundle
        :param relay_message_in: The incoming relay message
        :param relay_message_out: Thr outgoing relay message
        """


class SimpleOptionHandler(OptionHandler):
    """
    Standard handler for simple static options

    :param option: The option instance to add to the response
    :param append: Always add, even if an option of this class already exists
    :param always_send: Always send this option, even if the OptionRequestOption doesn't ask for it
    """

    def __init__(self, option: Option, *, append: bool = False, always_send: bool = False):
        self.option = option
        """The option instance to add to the response"""

        self.option_class = type(option)
        """The class of the option"""

        self.append = append
        """Always add, even if an option of this class already exists"""

        self.always_send = always_send
        """Always send this option, even if the :class:`.OptionRequestOption` doesn't ask for it"""

    def handle(self, bundle: TransactionBundle):
        """
        Add the option to the response in the bundle.

        :param bundle: The transaction bundle
        """
        # Make sure this option can go into this type of response
        if not bundle.response.may_contain(self.option):
            return

        # Check what the client requested
        if not self.always_send:
            # Don't add if the client doesn't request it
            oro = bundle.request.get_option_of_type(OptionRequestOption)
            if oro and self.option.option_type not in oro.requested_options:
                # Client doesn't want this
                return

        if self.append:
            # Just add
            add = True
        else:
            # See if this option was already present
            found = bundle.response.get_option_of_type(self.option_class)
            add = not found

        if add:
            # We always want to add it, or it didn't exist yet
            bundle.response.options.append(self.option)


class OverwritingOptionHandler(OptionHandler):
    """
    Overwriting handler for simple static options.

    :param option: The option instance to use
    :param always_send: Always send this option, even if the OptionRequestOption doesn't ask for it
    """

    def __init__(self, option: object, *, always_send: bool = False):
        """
        :type option: Option
        """
        self.option = option
        """The option to add to the response"""

        self.option_class = type(option)
        """The class of the option"""

        self.always_send = always_send
        """Whether an :class:`.OptionRequestOption` in the request should be ignored"""

    def handle(self, bundle: TransactionBundle):
        """
        Overwrite the option in the response in the bundle.

        :param bundle: The transaction bundle
        """
        # Make sure this option can go into this type of response
        if not bundle.response.may_contain(self.option):
            return

        # Check what the client requested
        if not self.always_send:
            # Don't add if the client doesn't request it
            oro = bundle.request.get_option_of_type(OptionRequestOption)
            if oro and self.option.option_type not in oro.requested_options:
                # Client doesn't want this
                return

        # Make sure this option isn't present and then add our own
        bundle.response.options = [existing_option for existing_option in bundle.response.options
                                   if not isinstance(existing_option, self.option_class)]
        bundle.response.options.insert(0, self.option)


class CopyOptionHandler(OptionHandler):
    """
    This handler just copies a type of option from the request to the response

    :param option_class: The option class to copy
    :param always_send: Always send this option, even if the OptionRequestOption doesn't ask for it
    """

    def __init__(self, option_class: object, *, always_send: bool = False):
        """
        :type option_class: Option
        """
        self.option_class = option_class
        """The class of the option from the request to the response"""

        self.always_send = always_send
        """Whether an :class:`.OptionRequestOption` in the request should be ignored"""

    def handle(self, bundle: TransactionBundle):
        """
        Copy the option from the request to the response.

        :param bundle: The transaction bundle
        """
        # Make sure this option can go into this type of response
        if not bundle.response.may_contain(self.option_class):
            return

        # Check what the client requested
        if not self.always_send:
            # Don't add if the client doesn't request it
            oro = bundle.request.get_option_of_type(OptionRequestOption)
            if oro and self.option_class.option_type not in oro.requested_options:
                # Client doesn't want this
                return

        # Make sure this option isn't present and then copy those from the request
        bundle.response.options = [existing_option for existing_option in bundle.response.options
                                   if not isinstance(existing_option, self.option_class)]
        bundle.response.options[:0] = bundle.request.get_options_of_type(self.option_class)


class CopyRelayOptionHandler(RelayOptionHandler):
    """
    This handler just copies a type of option from the incoming relay messages to the outgoing relay messages

    :param option_class: The option class to copy
    """

    def __init__(self, option_class: object):
        """
        :type option_class: Option
        """
        self.option_class = option_class
        """The class of the option from the :class:`.RelayForwardMessage` to the :class:`.RelayReplyMessage`"""

    def handle_relay(self, bundle: TransactionBundle,
                     relay_message_in: RelayForwardMessage, relay_message_out: RelayReplyMessage):
        """
        Copy the options for each relay message pair.

        :param bundle: The transaction bundle
        :param relay_message_in: The incoming relay message
        :param relay_message_out: Thr outgoing relay message
        """
        # Make sure this option can go into this type of response
        if not relay_message_out.may_contain(self.option_class):
            return

        # Make sure this option isn't present and then copy those from the request
        relay_message_out.options = [existing_option for existing_option in relay_message_out.options
                                     if not isinstance(existing_option, self.option_class)]
        relay_message_out.options[:0] = relay_message_in.get_options_of_type(self.option_class)



"""
Constants relevant for the IPv6 DHCP protocol
"""

All_DHCP_Relay_Agents_and_Servers = 'ff02::1:2'
All_DHCP_Servers = 'ff05::1:3'

CLIENT_PORT = 546
SERVER_PORT = 547

# Transmission and Retransmission Parameters
# :rfc:`3315#section-5.5`
SOL_MAX_DELAY = 1
SOL_TIMEOUT = 1
SOL_MAX_RT = 120
REQ_TIMEOUT = 1
REQ_MAX_RT = 30
REQ_MAX_RC = 10
CNF_MAX_DELAY = 1
CNF_TIMEOUT = 1
CNF_MAX_RT = 4
CNF_MAX_RD = 10
REN_TIMEOUT = 10
REN_MAX_RT = 600
REB_TIMEOUT = 10
REB_MAX_RT = 600
INF_MAX_DELAY = 1
INF_TIMEOUT = 1
INF_MAX_RT = 120
REL_TIMEOUT = 1
REL_MAX_RC = 5
DEC_TIMEOUT = 1
DEC_MAX_RC = 5
REC_TIMEOUT = 2
REC_MAX_RC = 8
HOP_COUNT_LIMIT = 32

# Representation of time values and "Infinity" as a time value
# :rfc:`3315#section-5.6`
INFINITY = 0xffffffff



"""
Option handlers that limit the t1/t2 values in replies
"""

import configparser
from abc import ABCMeta, abstractmethod

from dhcpkit.ipv6 import INFINITY
from dhcpkit.ipv6.extensions.prefix_delegation import IAPDOption, IAPrefixOption
from dhcpkit.ipv6.option_handlers import OptionHandler
from dhcpkit.ipv6.options import Option, IANAOption, IAAddressOption
from dhcpkit.ipv6.transaction_bundle import TransactionBundle


class TimingLimitsOptionHandler(OptionHandler, metaclass=ABCMeta):
    """
    A handler that limits the t1/t2 values in an option

    :type min_t1: int
    :type max_t1: int
    :type factor_t1: float or None
    :type min_t2: int
    :type max_t2: int
    :type factor_t2: float or None
    """

    def __init__(self,
                 min_t1=0, max_t1=INFINITY, factor_t1=0.5,
                 min_t2=0, max_t2=INFINITY, factor_t2=0.8):

        # These are the outer limits
        self.min_t1 = max(0, min_t1)
        self.max_t2 = min(max_t2, INFINITY)

        # If t2 has a max then t1's max must be at least as small
        self.max_t1 = min(max_t1, self.max_t2)

        # If t1 has a min then t2's min must be at least as large
        self.min_t2 = max(self.min_t1, min_t2)

        # Store the factors to auto-calculate t1/t2 based on the shortest preferred lifetime
        if factor_t1 is not None:
            self.factor_t1 = min(max(0, factor_t1), 1)
        else:
            self.factor_t1 = None

        if factor_t2 is not None:
            self.factor_t2 = min(max(0, factor_t2), 1)
        else:
            self.factor_t2 = None

        # Do some basic checks for impossible values
        if self.min_t1 > self.max_t2:
            raise ValueError("t1 must be able to be smaller than t2")

        if self.factor_t1 and self.factor_t2 and self.factor_t1 > self.factor_t2:
            raise ValueError("t1 factor must be smaller than t2 factor")

    @staticmethod
    def str_to_time(value: str) -> int:
        """
        Cast the string INFINITY to the infinity value, otherwise convert to an integer

        :param value: The string to parse
        :return: The integer value
        """
        if value.upper() == 'INFINITY':
            return INFINITY
        return int(value)

    @staticmethod
    def str_to_factor(value: str) -> float or None:
        """
        Cast the string NONE to the None value, otherwise convert to a float

        :param value: The string to parse
        :return: The float value or None
        """
        if value.upper() == 'NONE':
            return None
        return float(value)

    @classmethod
    def from_config(cls, section: configparser.SectionProxy, option_handler_id: str=None) -> OptionHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :param section: The configuration section
        :param option_handler_id: Optional extra identifier
        :return: A handler object
        :rtype: OptionHandler
        """
        min_t1 = cls.str_to_time(section.get('min-t1', '0'))
        max_t1 = cls.str_to_time(section.get('max-t1', 'INFINITY'))
        factor_t1 = cls.str_to_factor(section.get('factor-t1', '0.5'))

        min_t2 = cls.str_to_time(section.get('min-t2', '0'))
        max_t2 = cls.str_to_time(section.get('max-t2', 'INFINITY'))
        factor_t2 = cls.str_to_factor(section.get('factor-t2', '0.8'))

        return cls(min_t1, max_t1, factor_t1, min_t2, max_t2, factor_t2)

    @staticmethod
    @abstractmethod
    def filter_options(options: [Option]) -> [Option]:
        """
        Extract the options that we want to set the t1/t2 values of.

        :param options: The options in the response message
        :returns: The relevant options of the response message
        :rtype: list[IANAOption]
        """

    @staticmethod
    @abstractmethod
    def extract_preferred_lifetime(option: Option) -> int or None:
        """
        Extract the preferred lifetime from the given (sub)option. Returns None if this option doesn't contain a
        preferred lifetime.

        :param option: The option to extract the preferred lifetime from
        :returns: The preferred lifetime, if any
        """

    def handle(self, bundle: TransactionBundle):
        """
        Make sure the T1/T2 values are within the set limits.

        :param bundle: The transaction bundle
        """
        # Make a list of IAIDs in the response
        for option in self.filter_options(bundle.response.options):
            # Find the shortest preferred lifetime
            shortest_preferred = INFINITY + 1
            for suboption in option.options:
                preferred_lifetime = self.extract_preferred_lifetime(suboption)
                if preferred_lifetime is not None:
                    shortest_preferred = min(shortest_preferred, preferred_lifetime)

            # Don't mess with the timers if there are no addresses
            if shortest_preferred > INFINITY:
                continue

            # Calculate t1
            if option.t1 == 0 and self.factor_t1 is not None:
                if shortest_preferred == INFINITY:
                    option.t1 = INFINITY
                else:
                    option.t1 = int(shortest_preferred * self.factor_t1)

            # Calculate t2
            if option.t2 == 0 and self.factor_t2 is not None:
                if shortest_preferred == INFINITY:
                    option.t2 = INFINITY
                else:
                    option.t2 = int(shortest_preferred * self.factor_t2)

            # Now limit to the min/max bounds, making sure that t1 is not larger than t2
            # This will ignore the given boundaries if they conflict with shortest_preferred lifetime
            option.t2 = min(max(self.min_t2, option.t2), self.max_t2, shortest_preferred)
            option.t1 = min(max(self.min_t1, option.t1), self.max_t1, option.t2)


class IANATimingLimitsOptionHandler(TimingLimitsOptionHandler):
    """
    A handler that limits the t1/t2 values in an IANAOption
    """

    @staticmethod
    def filter_options(options: [Option]) -> [IANAOption]:
        """
        Extract the IANAOptions that we want to set the t1/t2 values of.

        :param options: The options in the response message
        :returns: The relevant options of the response message
        :rtype: list[IANAOption]
        """
        return [option for option in options if isinstance(option, IANAOption)]

    @staticmethod
    def extract_preferred_lifetime(option: Option) -> int or None:
        """
        Extract the preferred lifetime from the given (sub)option. Returns None if this option doesn't contain a
        preferred lifetime.

        :param option: The option to extract the preferred lifetime from
        :returns: The preferred lifetime, if any
        """
        if isinstance(option, IAAddressOption):
            return option.preferred_lifetime
        else:
            return None


class IAPDTimingLimitsOptionHandler(TimingLimitsOptionHandler):
    """
    A handler that limits the t1/t2 values in an IANAOption
    """

    @staticmethod
    def filter_options(options: [Option]) -> [IAPDOption]:
        """
        Extract the IAPDOptions that we want to set the t1/t2 values of.

        :param options: The options in the response message
        :returns: The relevant options of the response message
        :rtype: list[IAPDOption]
        """
        return [option for option in options if isinstance(option, IAPDOption)]

    @staticmethod
    def extract_preferred_lifetime(option: Option) -> int or None:
        """
        Extract the preferred lifetime from the given (sub)option. Returns None if this option doesn't contain a
        preferred lifetime.

        :param option: The option to extract the preferred lifetime from
        :returns: The preferred lifetime, if any
        """
        if isinstance(option, IAPrefixOption):
            return option.preferred_lifetime
        else:
            return None



"""
Utility functions for IPv6 DHCP
"""
from ipaddress import IPv6Address, IPv6Network


def address_in_prefixes(address: IPv6Address, prefixes: [IPv6Network]) -> bool:
    """
    Check whether the given address is part of one of the given prefixes

    :param address: The IPv6 address to check
    :param prefixes: The list of IPv6 prefixes
    :type prefixes: list[IPv6Network]
    :return: Whether the address is part of one of the prefixes
    """
    for prefix in prefixes:
        if address in prefix:
            return True

    return False


def prefix_overlaps_prefixes(prefix: IPv6Network, prefixes: [IPv6Network]) -> bool:
    """
    Check whether the given address is part of one of the given prefixes

    :param prefix: The IPv6 prefix to check
    :param prefixes: The list of IPv6 prefixes
    :type prefixes: list[IPv6Network]
    :return: Whether the address is part of one of the prefixes
    """
    for other_prefix in prefixes:
        if prefix.overlaps(other_prefix):
            return True

    return False



"""
Option handlers for the basic :rfc:`3315` options
"""

import configparser
import logging
from ipaddress import IPv6Address

from dhcpkit.ipv6.duids import DUID
from dhcpkit.ipv6.exceptions import CannotRespondError
from dhcpkit.ipv6.messages import ConfirmMessage, ReleaseMessage, DeclineMessage
from dhcpkit.ipv6.option_handlers import CopyOptionHandler, OverwritingOptionHandler, SimpleOptionHandler, \
    OptionHandler
from dhcpkit.ipv6.options import ClientIdOption, ServerIdOption, PreferenceOption, ServerUnicastOption, \
    StatusCodeOption, STATUS_SUCCESS
from dhcpkit.ipv6.transaction_bundle import TransactionBundle

logger = logging.getLogger(__name__)


class ClientIdOptionHandler(CopyOptionHandler):
    """
    The handler for ClientIdOptions
    """

    def __init__(self):
        super().__init__(ClientIdOption, always_send=True)


class ServerIdOptionHandler(OverwritingOptionHandler):
    """
    The handler for ServerIdOption. Checks whether any server-id in the request matches our own and puts our server-id
    in the response message to let the client know who is answering.

    :type option: ServerIdOption
    """

    option = None

    def __init__(self, duid: DUID):
        """
        Create a handler function based on the provided DUID

        :param duid: The DUID of this server
        """
        option = ServerIdOption(duid)
        option.validate()

        super().__init__(option, always_send=True)

    def pre(self, bundle: TransactionBundle):
        """
        Check if there is a ServerId in the request

        :param bundle: The transaction bundle
        """
        server_id = bundle.request.get_option_of_type(ServerIdOption)
        if server_id and server_id.duid != self.option.duid:
            # This message is not for this server
            raise CannotRespondError


class PreferenceOptionHandler(SimpleOptionHandler):
    """
    The handler for PreferenceOption which adds a preference option to appropriate responses
    """

    def __init__(self, preference: int):
        # This option remains constant, so create a singleton that can be re-used
        option = PreferenceOption(preference=preference)
        option.validate()

        super().__init__(option, always_send=True)

    @classmethod
    def from_config(cls, section: configparser.SectionProxy, option_handler_id: str = None) -> OptionHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :param section: The configuration section
        :param option_handler_id: Optional extra identifier
        :return: A handler object
        :rtype: OptionHandler
        """
        preference = section.getint('preference')
        if preference is None:
            raise configparser.NoOptionError('preference', section.name)

        return cls(preference)


class ServerUnicastOptionHandler(SimpleOptionHandler):
    """
    The handler for inserting ServerUniCastOptions into responses
    """

    def __init__(self, address: IPv6Address):
        # This option remains constant, so create a singleton that can be re-used
        option = ServerUnicastOption(server_address=address)
        option.validate()

        super().__init__(option, always_send=True)

    @classmethod
    def from_config(cls, section: configparser.SectionProxy, option_handler_id: str = None) -> OptionHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :param section: The configuration section
        :param option_handler_id: Optional extra identifier
        :return: A handler object
        :rtype: OptionHandler
        """
        address = section.get('server-address')
        if address is None:
            raise configparser.NoOptionError('server-address', section.name)

        address = IPv6Address(address)

        return cls(address)


class ConfirmStatusOptionHandler(OptionHandler):
    """
    The handler that makes sure that replies to confirm messages have a status code. When we reach the end without any
    status code being set we assume success. Other option handlers set the status to something else if they cannot
    confirm their part.
    """

    def handle(self, bundle: TransactionBundle):
        """
        Update the status of the reply to :class:`.ConfirmMessage`.

        :param bundle: The transaction bundle
        """
        if isinstance(bundle.request, ConfirmMessage):
            existing = bundle.response.get_option_of_type(StatusCodeOption)
            if not existing:
                bundle.response.options.append(
                    StatusCodeOption(STATUS_SUCCESS, "Assignments confirmed")
                )


class ReleaseStatusOptionHandler(OptionHandler):
    """
    The handler that makes sure that replies to release messages have a status code. When we reach the end without any
    status code being set we assume success. Other option handlers set the status to something else if they cannot
    confirm their part.
    """

    def handle(self, bundle: TransactionBundle):
        """
        Update the status of the reply to :class:`.ReleaseMessage`.

        :param bundle: The transaction bundle
        """
        if isinstance(bundle.request, ReleaseMessage):
            existing = bundle.response.get_option_of_type(StatusCodeOption)
            if not existing:
                bundle.response.options.append(
                    StatusCodeOption(STATUS_SUCCESS,
                                     "Thank you for releasing your resources")
                )


class DeclineStatusOptionHandler(OptionHandler):
    """
    The handler that makes sure that replies to decline messages have a status code. When we reach the end without any
    status code being set we assume success. Other option handlers set the status to something else if they cannot
    confirm their part.
    """

    def handle(self, bundle: TransactionBundle):
        """
        Update the status of the reply to :class:`.DeclineMessage`.

        :param bundle: The transaction bundle
        """
        if isinstance(bundle.request, DeclineMessage):
            existing = bundle.response.get_option_of_type(StatusCodeOption)
            if not existing:
                bundle.response.options.append(
                    StatusCodeOption(STATUS_SUCCESS,
                                     "Our apologies for assigning you unusable addresses")
                )



"""
Test whether the basic stuff of ProtocolElement works as intended
"""
import unittest

from dhcpkit.protocol_element import ProtocolElement


class DemoElementBase(ProtocolElement):
    """
    A simple element to test with
    """

    @classmethod
    def determine_class(cls, buffer: bytes, offset: int=0) -> type:
        """
        Intentionally left empty. Specific implementations must be tested separately.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :return: The best known class for this data
        """

    def load_from(self, buffer: bytes, offset: int=0, length: int=None) -> int:
        """
        Intentionally left empty. Specific implementations must be tested separately.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """

    def save(self) -> bytes:
        """
        Intentionally left empty. Specific implementations must be tested separately.

        :return: The buffer with the data from this element
        """


class DemoElement(DemoElementBase):
    """
    Sub-element to test with
    """


class OneParameterDemoElement(DemoElementBase):
    """
    Sub-element to test with
    """

    def __init__(self, one: DemoElementBase):
        self.one = one


class TwoParameterDemoElement(DemoElementBase):
    """
    Sub-element to test with
    """

    def __init__(self, one: int, two: DemoElementBase):
        self.one = one
        self.two = two


class ThreeParameterDemoElement(DemoElementBase):
    """
    Sub-element to test with
    """

    def __init__(self, one: int, two: str, three: [DemoElementBase]):
        self.one = one
        self.two = two
        self.three = three or []


class BadDemoElement(DemoElementBase):
    """
    Sub-element to test with
    """


class ContainerElementBase(DemoElementBase):
    """
    A simple element that contains DemoElements
    """

    def __init__(self, elements: [ProtocolElement]):
        self.elements = elements or []

    def validate(self):
        """
        Validate the contents of this element
        """
        self.validate_contains(self.elements)


class AnythingContainerElement(ContainerElementBase):
    """
    Container that may contain as many as it wants
    """


class NothingContainerElement(ContainerElementBase):
    """
    Container that may contain as many as it wants
    """


class MinOneContainerElement(ContainerElementBase):
    """
    Container that must contain at least one sub-element
    """


class MaxOneContainerElement(ContainerElementBase):
    """
    Container that must contain at most one sub-element
    """


class ExactlyOneContainerElement(ContainerElementBase):
    """
    Container that must contain exactly one sub-element
    """


class ExactlyTwoContainerElement(ContainerElementBase):
    """
    Container that must contain exactly two sub-elements
    """


AnythingContainerElement.add_may_contain(DemoElement)
NothingContainerElement.add_may_contain(DemoElement, 0, 0)
MinOneContainerElement.add_may_contain(DemoElement, 1)
MaxOneContainerElement.add_may_contain(DemoElement, 0, 1)
ExactlyOneContainerElement.add_may_contain(DemoElement, 1, 1)
ExactlyTwoContainerElement.add_may_contain(DemoElement, 2, 2)


class ElementOccurrenceTestCase(unittest.TestCase):
    def test_bad(self):
        container = AnythingContainerElement(elements=[BadDemoElement()])
        self.assertRaisesRegex(ValueError, 'cannot contain BadDemoElement', container.validate)

    def test_class_based(self):
        container = AnythingContainerElement(elements=[])
        self.assertTrue(container.may_contain(DemoElement))
        self.assertFalse(container.may_contain(BadDemoElement))

    def test_anything_0(self):
        container = AnythingContainerElement(elements=[])
        container.validate()

    def test_anything_1(self):
        container = AnythingContainerElement(elements=[DemoElement()])
        container.validate()

    def test_anything_2(self):
        container = AnythingContainerElement(elements=[DemoElement(), DemoElement()])
        container.validate()

    def test_nothing_0(self):
        container = NothingContainerElement(elements=[])
        container.validate()

    def test_nothing_1(self):
        container = NothingContainerElement(elements=[DemoElement()])
        self.assertRaisesRegex(ValueError, 'cannot contain DemoElement', container.validate)

    def test_nothing_2(self):
        container = NothingContainerElement(elements=[DemoElement(), DemoElement()])
        self.assertRaisesRegex(ValueError, 'cannot contain DemoElement', container.validate)

    def test_min_one_0(self):
        container = MinOneContainerElement(elements=[])
        self.assertRaisesRegex(ValueError, 'must contain at least 1 DemoElement', container.validate)

    def test_min_one_1(self):
        container = MinOneContainerElement(elements=[DemoElement()])
        container.validate()

    def test_min_one_2(self):
        container = MinOneContainerElement(elements=[DemoElement(), DemoElement()])
        container.validate()

    def test_max_one_0(self):
        container = MaxOneContainerElement(elements=[])
        container.validate()

    def test_max_one_1(self):
        container = MaxOneContainerElement(elements=[DemoElement()])
        container.validate()

    def test_max_one_2(self):
        container = MaxOneContainerElement(elements=[DemoElement(), DemoElement()])
        self.assertRaisesRegex(ValueError, 'may only contain 1 DemoElement', container.validate)

    def test_exactly_one_0(self):
        container = ExactlyOneContainerElement(elements=[])
        self.assertRaisesRegex(ValueError, 'must contain at least 1 DemoElement', container.validate)

    def test_exactly_one_1(self):
        container = ExactlyOneContainerElement(elements=[DemoElement()])
        container.validate()

    def test_exactly_one_2(self):
        container = ExactlyOneContainerElement(elements=[DemoElement(), DemoElement()])
        self.assertRaisesRegex(ValueError, 'only contain 1', container.validate)

    def test_exactly_two_1(self):
        container = ExactlyTwoContainerElement(elements=[DemoElement()])
        self.assertRaisesRegex(ValueError, 'must contain at least 2 DemoElements', container.validate)

    def test_exactly_two_2(self):
        container = ExactlyTwoContainerElement(elements=[DemoElement(), DemoElement()])
        container.validate()

    def test_exactly_two_3(self):
        container = ExactlyTwoContainerElement(elements=[DemoElement(), DemoElement(), DemoElement()])
        self.assertRaisesRegex(ValueError, 'may only contain 2 DemoElements', container.validate)

    def test_compare(self):
        container1 = AnythingContainerElement(elements=[DemoElement(), DemoElement()])
        container2 = AnythingContainerElement(elements=[DemoElement(), DemoElement()])
        container3 = AnythingContainerElement(elements=[DemoElement(), BadDemoElement()])

        self.assertEqual(container1, container2)
        self.assertNotEqual(container1, container3)

    def test_repr(self):
        element = TwoParameterDemoElement(1608, DemoElement())
        self.assertEqual(repr(element), "TwoParameterDemoElement(one=1608, two=DemoElement())")

    def test_str_no_parameters(self):
        element = DemoElement()
        self.assertEqual(str(element), "DemoElement()")

    def test_str_one_parameter(self):
        element = OneParameterDemoElement(DemoElement())
        self.assertEqual(str(element), "OneParameterDemoElement(one=DemoElement())")

        element = OneParameterDemoElement(one=DemoElement())
        self.assertEqual(str(element), "OneParameterDemoElement(one=DemoElement())")

        element = OneParameterDemoElement(one=TwoParameterDemoElement(1608, DemoElement()))
        self.assertEqual(str(element), "OneParameterDemoElement(\n"
                                       "  one=TwoParameterDemoElement(\n"
                                       "    one=1608,\n"
                                       "    two=DemoElement(),\n"
                                       "  )\n"
                                       ")")

    def test_str_two_parameters(self):
        element = TwoParameterDemoElement(1608, DemoElement())
        self.assertEqual(str(element), "TwoParameterDemoElement(\n"
                                       "  one=1608,\n"
                                       "  two=DemoElement(),\n"
                                       ")")

    def test_str_three_parameters(self):
        element = ThreeParameterDemoElement(1608, 'Something', [
            DemoElement(),
            OneParameterDemoElement(DemoElement()),
            TwoParameterDemoElement(1608, TwoParameterDemoElement(1608, DemoElement()))
        ])
        self.assertEqual(str(element), "ThreeParameterDemoElement(\n"
                                       "  one=1608,\n"
                                       "  two='Something',\n"
                                       "  three=[\n"
                                       "    DemoElement(),\n"
                                       "    OneParameterDemoElement(one=DemoElement()),\n"
                                       "    TwoParameterDemoElement(\n"
                                       "      one=1608,\n"
                                       "      two=TwoParameterDemoElement(\n"
                                       "        one=1608,\n"
                                       "        two=DemoElement(),\n"
                                       "      ),\n"
                                       "    ),\n"
                                       "  ],\n"
                                       ")")


if __name__ == '__main__':
    unittest.main()



"""
There is not much to test with these simple exceptions. This test checks if the inheritance hierarchy is as expected.
"""
import unittest

from dhcpkit.ipv6.exceptions import HandlerException, CannotRespondError, UseMulticastError, ListeningSocketError, \
    InvalidPacketError


class ExceptionHierarchyTestCase(unittest.TestCase):
    def test_handler_exception(self):
        with self.assertRaises(Exception):
            raise HandlerException

        with self.assertRaises(HandlerException):
            raise HandlerException

    def test_cannot_respond_error(self):
        with self.assertRaises(Exception):
            raise CannotRespondError

        with self.assertRaises(HandlerException):
            raise CannotRespondError

        with self.assertRaises(CannotRespondError):
            raise CannotRespondError

    def test_use_multicast_error(self):
        with self.assertRaises(Exception):
            raise UseMulticastError

        with self.assertRaises(HandlerException):
            raise UseMulticastError

        with self.assertRaises(UseMulticastError):
            raise UseMulticastError

    def test_listening_socket_error(self):
        with self.assertRaises(Exception):
            raise ListeningSocketError

        with self.assertRaises(ListeningSocketError):
            raise ListeningSocketError

    def test_invalid_packet_error(self):
        with self.assertRaises(Exception):
            raise InvalidPacketError

        with self.assertRaisesRegex(InvalidPacketError, r"Invalid packet from \('2001:db8::1', 546, 0, 1\)"):
            raise InvalidPacketError(sender=('2001:db8::1', 546, 0, 1))


if __name__ == '__main__':
    unittest.main()



"""
Implementation of SNTP option as specified in :rfc:`4075`.
"""

from ipaddress import IPv6Address
from struct import pack

from dhcpkit.ipv6.options import Option

OPTION_SNTP_SERVERS = 31


class SNTPServersOption(Option):
    """
    :rfc:`4075#section-4`

    The Simple Network Time Protocol servers option provides a list of
    one or more IPv6 addresses of SNTP [3] servers available to the
    client for synchronization.  The clients use these SNTP servers to
    synchronize their system time to that of the standard time servers.
    Clients MUST treat the list of SNTP servers as an ordered list.  The
    server MAY list the SNTP servers in decreasing order of preference.

    The option defined in this document can only be used to configure
    information about SNTP servers that can be reached using IPv6.  The
    DHCP option to configure information about IPv4 SNTP servers can be
    found in :rfc:`2132` [4].  Mechanisms for configuring IPv4/IPv6 dual-
    stack applications are being considered, but are not specified in
    this document.

    The format of the Simple Network Time Protocol servers option is as
    shown below::

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |      OPTION_SNTP_SERVERS       |        option-len            |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                               |
      |                  SNTP server (IPv6 address)                   |
      |                                                               |
      |                                                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                               |
      |                  SNTP server (IPv6 address)                   |
      |                                                               |
      |                                                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                              ...                              |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    option-code
        OPTION_SNTP_SERVERS (31).

    option-len
        Length of the 'SNTP server' fields, in octets; it must be a multiple of 16.

    SNTP server
        IPv6 address of SNTP server.
    """

    option_type = OPTION_SNTP_SERVERS

    def __init__(self, sntp_servers: [IPv6Address] = None):
        self.sntp_servers = sntp_servers or []
        """List of IPv6 addresses of SNTP servers"""

    def validate(self):
        """
        Validate that the contents of this object conform to protocol specs.
        """
        if not isinstance(self.sntp_servers, list):
            raise ValueError("SNTP servers must be a list")

        for address in self.sntp_servers:
            if not isinstance(address, IPv6Address) or \
                    address.is_link_local or \
                    address.is_loopback or \
                    address.is_multicast or \
                    address.is_unspecified:
                raise ValueError("SNTP servers must be a list of routable IPv6 addresses")

    def load_from(self, buffer: bytes, offset: int = 0, length: int = None) -> int:
        """
        Load the internal state of this object from the given buffer. The buffer may contain more data after the
        structured element is parsed. This data is ignored.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """
        my_offset, option_len = self.parse_option_header(buffer, offset, length)
        header_offset = my_offset

        if option_len % 16 != 0:
            raise ValueError('SNTP Servers Option length must be a multiple of 16')

        # Parse the addresses
        self.sntp_servers = []
        max_offset = option_len + header_offset  # The option_len field counts bytes *after* the header fields
        while max_offset > my_offset:
            address = IPv6Address(buffer[offset + my_offset:offset + my_offset + 16])
            self.sntp_servers.append(address)
            my_offset += 16

        if my_offset != max_offset:
            raise ValueError('Option length does not match the combined length of the included addresses')

        self.validate()

        return my_offset

    def save(self) -> bytes:
        """
        Save the internal state of this object as a buffer.

        :return: The buffer with the data from this element
        """
        self.validate()

        buffer = bytearray()
        buffer.extend(pack('!HH', self.option_type, len(self.sntp_servers) * 16))
        for address in self.sntp_servers:
            buffer.extend(address.packed)

        return buffer



"""
Test the ElapsedTimeOption implementation
"""
import unittest

from dhcpkit.ipv6.options import ElapsedTimeOption
from tests.ipv6.options import test_option


class OptionRequestOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('0008000204d2')
        self.option_object = ElapsedTimeOption(elapsed_time=1234)
        self.parse_option()

    def test_validate_elapsed_time(self):
        self.option.elapsed_time = 0.1
        with self.assertRaisesRegex(ValueError, 'unsigned 16 bit integer'):
            self.option.validate()

        self.option.elapsed_time = 0
        self.option.validate()

        self.option.elapsed_time = -1
        with self.assertRaisesRegex(ValueError, 'unsigned 16 bit integer'):
            self.option.validate()

        self.option.elapsed_time = 65535
        self.option.validate()

        self.option.elapsed_time = 65536
        with self.assertRaisesRegex(ValueError, 'unsigned 16 bit integer'):
            self.option.validate()

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'must have length 2'):
            ElapsedTimeOption.parse(bytes.fromhex('000800000000ffff'))


if __name__ == '__main__':
    unittest.main()



"""
Implementation of Remote-ID option as specified in :rfc:`4649`.
"""

from struct import pack, unpack_from

from dhcpkit.ipv6.messages import RelayForwardMessage, RelayReplyMessage
from dhcpkit.ipv6.options import Option

OPTION_REMOTE_ID = 37


class RemoteIdOption(Option):
    """
    :rfc:`4649#section-3`

    This option may be added by DHCPv6 relay agents that terminate
    switched or permanent circuits and have mechanisms to identify the
    remote host end of the circuit.

    The format of the DHCPv6 Relay Agent Remote-ID option is shown below::

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |       OPTION_REMOTE_ID        |         option-len            |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                       enterprise-number                       |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      .                                                               .
      .                           remote-id                           .
      .                                                               .
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    option-code
        OPTION_REMOTE_ID (37).

    option-len
        4 + the length, in octets, of the remote-id field.  The minimum option-len is 5 octets.

    enterprise-number
        The vendor's registered Enterprise Number as registered with IANA [5].

    remote-id
        The opaque value for the remote-id.

    The definition of the remote-id carried in this option is vendor
    specific.  The vendor is indicated in the enterprise-number field.
    The remote-id field may be used to encode, for instance:

    - a "caller ID" telephone number for dial-up connection
    - a "user name" prompted for by a Remote Access Server
    - a remote caller ATM address
    - a "modem ID" of a cable data modem
    - the remote IP address of a point-to-point link
    - a remote X.25 address for X.25 connections
    - an interface or port identifier

    Each vendor must ensure that the remote-id is unique for its
    enterprise-number, as the octet sequence of enterprise-number
    followed by remote-id must be globally unique.  One way to achieve
    uniqueness might be to include the relay agent's DHCP Unique
    Identifier (DUID) [1] in the remote-id.

    :type enterprise_number: int
    :type remote_id: bytes
    """

    option_type = OPTION_REMOTE_ID

    def __init__(self, enterprise_number: int=0, remote_id: bytes=b''):
        self.enterprise_number = enterprise_number
        """The `enterprise number <http://www.iana.org/assignments/enterprise-numbers>`_ as registered with IANA"""

        self.remote_id = remote_id
        """The remote-id as bytes"""

    def validate(self):
        """
        Validate that the contents of this object conform to protocol specs.
        """
        if not isinstance(self.enterprise_number, int) or not (0 <= self.enterprise_number < 2 ** 32):
            raise ValueError("Enterprise number must be an unsigned 32 bit integer")

        if not isinstance(self.remote_id, bytes) or len(self.remote_id) >= 2 ** 16:
            raise ValueError("Remote-ID must be sequence of bytes")

    def load_from(self, buffer: bytes, offset: int=0, length: int=None) -> int:
        """
        Load the internal state of this object from the given buffer. The buffer may contain more data after the
        structured element is parsed. This data is ignored.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """
        my_offset, option_len = self.parse_option_header(buffer, offset, length)

        self.enterprise_number = unpack_from('!I', buffer, offset=offset + my_offset)[0]
        my_offset += 4

        remote_id_length = option_len - 4
        self.remote_id = buffer[offset + my_offset:offset + my_offset + remote_id_length]
        my_offset += remote_id_length

        self.validate()

        return my_offset

    def save(self) -> bytes:
        """
        Save the internal state of this object as a buffer.

        :return: The buffer with the data from this element
        """
        self.validate()
        return pack('!HHI', self.option_type, len(self.remote_id) + 4, self.enterprise_number) + self.remote_id


RelayForwardMessage.add_may_contain(RemoteIdOption)

# The RFC says there is no requirement for servers to include this option in replies, but it is not forbidden
RelayReplyMessage.add_may_contain(RemoteIdOption)



"""
All exceptions used when handling IPv6 DHCP
"""


class HandlerException(Exception):
    """
    Base class for handler exceptions
    """


class CannotRespondError(HandlerException):
    """
    This exception signals that we cannot reply to this client.
    """


class UseMulticastError(HandlerException):
    """
    This exception signals that a STATUS_USEMULTICAST should be returned to the client.
    """


class ListeningSocketError(Exception):
    """
    Signal that the listening socket could not be created.
    """


class InvalidPacketError(Exception):
    """
    Signal that an incoming message was invalid

    :type sender: (str, int, int, int)
    """

    def __init__(self, *args, sender=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.sender = sender

    def __str__(self):
        return "Invalid packet from {}".format(self.sender)



"""
Test the IANAOption implementation
"""
from ipaddress import IPv6Address
import unittest

from dhcpkit.ipv6.options import IANAOption, StatusCodeOption, STATUS_SUCCESS, UnknownOption, IAAddressOption
from tests.ipv6.options import test_option


class IANAOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('0003'  # option_type: OPTION_IA_NA
                                          '0044'  # option_length
                                          '41424344'  # iaid: ABCD
                                          '00000029'  # t1: 41
                                          '0000002a'  # t2: 42
                                          '0005'  # option_type: OPTION_IAADDR
                                          '0018'  # option_length
                                          '20010db8000000000000000000000001'  # address: 2001:db8::1
                                          '00000000'  # preferred_lifetime
                                          '00000000'  # valid_lifetime
                                          '000d'  # option_type: OPTION_STATUS_CODE
                                          '0018'  # option_length
                                          '0000'  # status_code
                                          '45766572797468696e6720697320617765736f6d6521')  # status_message
        self.option_object = IANAOption(iaid=b'ABCD', t1=41, t2=42, options=[
            IAAddressOption(address=IPv6Address('2001:db8::1')),
            StatusCodeOption(status_code=STATUS_SUCCESS, status_message='Everything is awesome!')
        ])
        self.parse_option()

    def test_validate_iaid(self):
        self.option.iaid = b'ABC'
        with self.assertRaisesRegex(ValueError, 'must be four bytes'):
            self.option.validate()

        self.option.iaid = b'ABCDE'
        with self.assertRaisesRegex(ValueError, 'must be four bytes'):
            self.option.validate()

        self.option.iaid = 'ABCD'
        with self.assertRaisesRegex(ValueError, 'must be four bytes'):
            self.option.validate()

    def test_validate_t1(self):
        self.option.t1 = 0.1
        with self.assertRaisesRegex(ValueError, 'unsigned 32 bit integer'):
            self.option.validate()

        self.option.t1 = -1
        with self.assertRaisesRegex(ValueError, 'unsigned 32 bit integer'):
            self.option.validate()

        self.option.t1 = 2 ** 32
        with self.assertRaisesRegex(ValueError, 'unsigned 32 bit integer'):
            self.option.validate()

    def test_validate_t2(self):
        self.option.t2 = 0.1
        with self.assertRaisesRegex(ValueError, 'unsigned 32 bit integer'):
            self.option.validate()

        self.option.t2 = -1
        with self.assertRaisesRegex(ValueError, 'unsigned 32 bit integer'):
            self.option.validate()

        self.option.t2 = 2 ** 32
        with self.assertRaisesRegex(ValueError, 'unsigned 32 bit integer'):
            self.option.validate()

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'length does not match'):
            IANAOption.parse(bytes.fromhex('0003000041424344000000290000002a'))

    def test_sort(self):
        self.assertFalse(self.option > self.option)
        self.assertTrue(self.option <= self.option)

        with self.assertRaises(TypeError):
            self.assertFalse(self.option > 0)

    def test_get_options_of_type(self):
        found_options = self.option.get_options_of_type(StatusCodeOption)
        self.assertEqual(len(found_options), 1)
        self.assertIsInstance(found_options[0], StatusCodeOption)

        # But our test-cases don't have an UnknownOption in them
        found_options = self.option.get_options_of_type(UnknownOption)
        self.assertEqual(len(found_options), 0)

    def test_get_option_of_type(self):
        found_option = self.option.get_option_of_type(StatusCodeOption)
        self.assertIsInstance(found_option, StatusCodeOption)

        # But our test-cases don't have an UnknownOption in them
        found_option = self.option.get_option_of_type(UnknownOption)
        self.assertIsNone(found_option)

    def test_get_addresses(self):
        addresses = self.option.get_addresses()
        self.assertListEqual(addresses, [IPv6Address('2001:db8::1')])


if __name__ == '__main__':
    unittest.main()



"""
Option handlers for the DNS options defined in dhcpkit.ipv6.extensions.ntp
"""

import configparser
import re

from dhcpkit.ipv6.extensions.ntp import NTPSubOption, NTPServersOption, name_registry
from dhcpkit.ipv6.option_handlers import SimpleOptionHandler, OptionHandler
from dhcpkit.utils import camelcase_to_dash


class NTPServersOptionHandler(SimpleOptionHandler):
    """
    Handler for putting RecursiveNameServersOption in responses
    """

    def __init__(self, sub_options: [NTPSubOption]):
        option = NTPServersOption(options=sub_options)
        option.validate()

        super().__init__(option)

    @classmethod
    def from_config(cls, section: configparser.SectionProxy, option_handler_id: str=None) -> OptionHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :param section: The configuration section
        :param option_handler_id: Optional extra identifier
        :return: A handler object
        :rtype: OptionHandler
        """
        sub_options = []

        for name, value in section.items():
            # Strip numbers from the end, this can be used to supply the same option multiple times
            name = name.rstrip('0123456789-')

            if '-' in name or '_' in name:
                suboption_name = name.replace('_', '-').lower()
            else:
                suboption_name = camelcase_to_dash(name)

            suboption = name_registry.get(suboption_name)
            if not suboption:
                raise configparser.ParsingError("Unknown suboption: {}".format(suboption_name))

            for suboption_value in re.split('[,\t ]+', value):
                if not suboption_value:
                    raise configparser.ParsingError("{} option has no value".format(name))

                sub_options.append(suboption.from_string(suboption_value))

        return cls(sub_options)



"""
Implementation of NTP options as specified in :rfc:`5908`.
"""

import configparser
from ipaddress import IPv6Address
from struct import unpack_from, pack

from dhcpkit.ipv6.messages import ClientServerMessage
from dhcpkit.ipv6.options import Option
from dhcpkit.protocol_element import ProtocolElement
from dhcpkit.utils import camelcase_to_dash, parse_domain_bytes, encode_domain

OPTION_NTP_SERVER = 56

NTP_SUBOPTION_SRV_ADDR = 1
NTP_SUBOPTION_MC_ADDR = 2
NTP_SUBOPTION_SRV_FQDN = 3

# Registry
# type: {int: Option}
registry = {}

# Name Registry
# type: {str: Option}
name_registry = {}


def register(subclass: type):
    """
    Register a new option type in the option registry.

    :param subclass: A subclass of Option that implements the option
    """
    if not issubclass(subclass, NTPSubOption):
        raise TypeError('Only NTPSubOptions can be registered')

    # Store based on number
    # noinspection PyUnresolvedReferences
    registry[subclass.suboption_type] = subclass

    # Store based on name
    name = subclass.__name__
    if name.startswith('NTP'):
        name = name[3:]
    if name.endswith('SubOption'):
        name = name[:-9]
    name = camelcase_to_dash(name)
    name_registry[name] = subclass


# This subclass remains abstract
# noinspection PyAbstractClass
class NTPSubOption(ProtocolElement):
    """
    :rfc:`5908`

    :type suboption_type: int
    """

    # This needs to be overwritten in subclasses
    suboption_type = 0

    @classmethod
    def from_string(cls, config: str) -> object:
        """
        Create this suboption based on the provided string

        :param config: The input string
        :return: The suboption object
        """
        raise configparser.Error("{} does not support loading from string".format(cls.__name__))

    @classmethod
    def determine_class(cls, buffer: bytes, offset: int = 0) -> type:
        """
        Return the appropriate subclass from the registry, or UnknownNTPSubOption if no subclass is registered.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :return: The best known class for this suboption data
        """
        suboption_type = unpack_from('!H', buffer, offset=offset)[0]
        return registry.get(suboption_type, UnknownNTPSubOption)

    def parse_suboption_header(self, buffer: bytes, offset: int = 0, length: int = None) -> (int, int):
        """
        Parse the option code and length from the buffer and perform some basic validation.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer and the value of the suboption-len field
        """
        suboption_type, suboption_len = unpack_from('!HH', buffer, offset=offset)
        my_offset = 4

        if suboption_type != self.suboption_type:
            raise ValueError('The provided buffer does not contain {} data'.format(self.__class__.__name__))

        if length is not None and suboption_len + my_offset > length:
            raise ValueError('This suboption is longer than the available buffer')

        return my_offset, suboption_len


class UnknownNTPSubOption(NTPSubOption):
    """
    Container for raw NTP sub-option content for cases where we don't know how to decode it.

    :type suboption_data: bytes
    """

    def __init__(self, suboption_type: int = 0, suboption_data: bytes = b''):
        self.suboption_type = suboption_type
        """Type code for this sub-option"""

        self.suboption_data = suboption_data
        """Data for this sub-option"""

    def validate(self):
        """
        Validate that the contents of this object conform to protocol specs.
        """
        if not isinstance(self.suboption_type, int) and not (0 <= self.suboption_type < 2 ** 16):
            raise ValueError("Sub-option type must be an unsigned 16 bit integer")

        if not isinstance(self.suboption_data, bytes):
            raise ValueError("Sub-option data must be sequence of bytes")

    def load_from(self, buffer: bytes, offset: int = 0, length: int = None) -> int:
        """
        Load the internal state of this object from the given buffer. The buffer may contain more data after the
        structured element is parsed. This data is ignored.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """
        my_offset = 0

        self.suboption_type, option_len = unpack_from('!HH', buffer, offset=offset + my_offset)
        my_offset += 4

        max_length = length or (len(buffer) - offset)
        if my_offset + option_len > max_length:
            raise ValueError('This suboption is longer than the available buffer')

        self.suboption_data = buffer[offset + my_offset:offset + my_offset + option_len]
        my_offset += option_len

        return my_offset

    def save(self) -> bytes:
        """
        Save the internal state of this object as a buffer.

        :return: The buffer with the data from this element
        """
        return pack('!HH', self.suboption_type, len(self.suboption_data)) + self.suboption_data


class NTPServerAddressSubOption(NTPSubOption):
    """
    :rfc:`5908#section-4.1`

    This suboption is intended to appear inside the OPTION_NTP_SERVER
    option.  It specifies the IPv6 unicast address of an NTP server or
    SNTP server available to the client.

    The format of the NTP Server Address Suboption is::

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |    NTP_SUBOPTION_SRV_ADDR     |        suboption-len = 16     |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                               |
      |                                                               |
      |                   IPv6 address of NTP server                  |
      |                                                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    IPv6 address of the NTP server
        An IPv6 address.

    suboption-code
        NTP_SUBOPTION_SRV_ADDR (1).

    suboption-len
        16.

    :type address: IPv6Address
    """

    suboption_type = NTP_SUBOPTION_SRV_ADDR

    def __init__(self, address: IPv6Address = None):
        self.address = address
        """IPv6 address of an NTP server"""

    def validate(self):
        """
        Validate that the contents of this object conform to protocol specs.
        """
        if not isinstance(self.address, IPv6Address) or self.address.is_link_local or self.address.is_loopback \
                or self.address.is_multicast or self.address.is_unspecified:
            raise ValueError("NTP server address must be a routable IPv6 address")

    @classmethod
    def from_string(cls, config: str) -> object:
        """
        Create this suboption based on the provided string, which must contain an IPv6 unicast address.

        :param config: The input string
        :return: The suboption object
        """
        address = IPv6Address(config)

        option = cls(address=address)
        option.validate()
        return option

    def load_from(self, buffer: bytes, offset: int = 0, length: int = None) -> int:
        """
        Load the internal state of this object from the given buffer. The buffer may contain more data after the
        structured element is parsed. This data is ignored.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """
        my_offset, suboption_len = self.parse_suboption_header(buffer, offset, length)

        if suboption_len != 16:
            raise ValueError('NTP Server Address SubOptions must have length 16')

        self.address = IPv6Address(buffer[offset + my_offset:offset + my_offset + 16])
        my_offset += 16

        return my_offset

    def save(self) -> bytes:
        """
        Save the internal state of this object as a buffer.

        :return: The buffer with the data from this element
        """
        buffer = bytearray()
        buffer.extend(pack('!HH', self.suboption_type, 16))
        buffer.extend(self.address.packed)
        return buffer


class NTPMulticastAddressSubOption(NTPSubOption):
    """
    :rfc:`5908#section-4.2`

    This suboption is intended to appear inside the OPTION_NTP_SERVER
    option.  It specifies the IPv6 address of the IPv6 multicast group
    address used by NTP on the local network.

    The format of the NTP Multicast Address Suboption is::

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |    NTP_SUBOPTION_MC_ADDR      |        suboption-len = 16     |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                               |
      |                                                               |
      |                   Multicast IPv6 address                      |
      |                                                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    Multicast IPv6 address
        An IPv6 address.

    suboption-code
        NTP_SUBOPTION_MC_ADDR (2).

    suboption-len
        16.

    :type address: IPv6Address
    """

    suboption_type = NTP_SUBOPTION_MC_ADDR

    def __init__(self, address: IPv6Address = None):
        self.address = address
        """IPv6 multicast group address"""

    def validate(self):
        """
        Validate that the contents of this object conform to protocol specs.
        """
        if not isinstance(self.address, IPv6Address) or not self.address.is_multicast:
            raise ValueError("NTP multicast address must be a multicast IPv6 address")

    @classmethod
    def from_string(cls, config: str) -> object:
        """
        Create this suboption based on the provided string, which must contain an IPv6 multicast address.

        :param config: The input string
        :return: The suboption object
        """
        address = IPv6Address(config)

        option = cls(address=address)
        option.validate()
        return option

    def load_from(self, buffer: bytes, offset: int = 0, length: int = None) -> int:
        """
        Load the internal state of this object from the given buffer. The buffer may contain more data after the
        structured element is parsed. This data is ignored.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """
        my_offset, suboption_len = self.parse_suboption_header(buffer, offset, length)

        if suboption_len != 16:
            raise ValueError('NTP Multicast Address SubOptions must have length 16')

        self.address = IPv6Address(buffer[offset + my_offset:offset + my_offset + 16])
        my_offset += 16

        return my_offset

    def save(self) -> bytes:
        """
        Save the internal state of this object as a buffer.

        :return: The buffer with the data from this element
        """
        buffer = bytearray()
        buffer.extend(pack('!HH', self.suboption_type, 16))
        buffer.extend(self.address.packed)
        return buffer


class NTPServerFQDNSubOption(NTPSubOption):
    """
    :rfc:`5908#section-4.3`

    This suboption is intended to appear inside the OPTION_NTP_SERVER
    option.  It specifies the FQDN of an NTP server or SNTP server
    available to the client.

    The format of the NTP Server FQDN Suboption is::

      0                   1                   2                   3
      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |    NTP_SUBOPTION_SRV_FQDN     |         suboption-len         |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                               |
      |                      FQDN of NTP server                       |
      :                                                               :
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    suboption-code
        NTP_SUBOPTION_SRV_FQDN (3).

    suboption-len
        Length of the included FQDN field.

    FQDN
        Fully-Qualified Domain Name of the NTP server or SNTP server. This field MUST be encoded as described in
        :rfc:`3315`, Section 8.  Internationalized domain names are not allowed in this field.

    :type fqdn: str
    """

    suboption_type = NTP_SUBOPTION_SRV_FQDN

    def __init__(self, fqdn: str = ''):
        self.fqdn = fqdn
        """Domain name of an NTP server"""

    def validate(self):
        """
        Validate that the contents of this object conform to protocol specs.
        """
        if len(self.fqdn) > 255:
            raise ValueError("NTP server FQDN must be 255 characters or less")

        if any([0 >= len(label) > 63 for label in self.fqdn.split('.')]):
            raise ValueError("NTP server FQDN domain labels must be 1 to 63 characters long")

    @classmethod
    def from_string(cls, config: str) -> object:
        """
        Create this suboption based on the provided string, which must contain a hostname.

        :param config: The input string
        :return: The suboption object
        """
        option = cls(fqdn=config)
        option.validate()
        return option

    def load_from(self, buffer: bytes, offset: int = 0, length: int = None) -> int:
        """
        Load the internal state of this object from the given buffer. The buffer may contain more data after the
        structured element is parsed. This data is ignored.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """
        my_offset, suboption_len = self.parse_suboption_header(buffer, offset, length)
        header_offset = my_offset

        # Parse the domain labels
        max_offset = suboption_len + header_offset  # The option_len field counts bytes *after* the header fields
        domain_name_len, self.fqdn = parse_domain_bytes(buffer, offset=offset + my_offset, length=suboption_len)
        my_offset += domain_name_len

        if my_offset != max_offset:
            raise ValueError('Option length does not match the length of the included fqdn')

        return my_offset

    def save(self) -> bytes:
        """
        Save the internal state of this object as a buffer.

        :return: The buffer with the data from this element
        """
        fqdn_buffer = encode_domain(self.fqdn)

        buffer = bytearray()
        buffer.extend(pack('!HH', self.suboption_type, len(fqdn_buffer)))
        buffer.extend(fqdn_buffer)
        return buffer


class NTPServersOption(Option):
    """
    :rfc:`5908#section-4`

    This option serves as a container for server location information
    related to one NTP server or Simple Network Time Protocol (SNTP)
    :rfc:`4330` server.  This option can appear multiple times in a DHCPv6
    message.  Each instance of this option is to be considered by the NTP
    client or SNTP client as a server to include in its configuration.

    The option itself does not contain any value.  Instead, it contains
    one or several suboptions that carry NTP server or SNTP server
    location.  This option MUST include one, and only one, time source
    suboption.  The currently defined time source suboptions are
    NTP_OPTION_SRV_ADDR, NTP_OPTION_SRV_MC_ADDR, and NTP_OPTION_SRV_FQDN.
    It carries the NTP server or SNTP server location as a unicast or
    multicast IPv6 address or as an NTP server or SNTP server FQDN.  More
    time source suboptions may be defined in the future.  While the FQDN
    option offers the most deployment flexibility, resiliency as well as
    security, the IP address options are defined to cover cases where a
    DNS dependency is not desirable.

    If the NTP server or SNTP server location is an IPv6 multicast
    address, the client SHOULD use this address as an NTP multicast group
    address and listen to messages sent to this group in order to
    synchronize its clock.

    The format of the NTP Server Option is::

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |      OPTION_NTP_SERVER        |          option-len           |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                         suboption-1                           |
      :                                                               :
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                         suboption-2                           |
      :                                                               :
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      :                                                               :
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                         suboption-n                           |
      :                                                               :
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    option-code
        OPTION_NTP_SERVER (56).

    option-len
        Total length of the included suboptions.

    This document does not define any priority relationship between the
    client's embedded configuration (if any) and the NTP or SNTP servers
    discovered via this option.  In particular, the client is allowed to
    simultaneously use its own configured NTP servers or SNTP servers and
    the servers discovered via DHCP.

    :type options: list[NTPSubOption]
    """

    option_type = OPTION_NTP_SERVER

    def __init__(self, options: [NTPSubOption] = None):
        self.options = options or []
        """List of NTP server sub-options"""

    def validate(self):
        """
        Validate that the contents of this object conform to protocol specs.
        """
        # Check if all options are allowed
        self.validate_contains(self.options)
        for option in self.options:
            option.validate()

    def load_from(self, buffer: bytes, offset: int = 0, length: int = None) -> int:
        """
        Load the internal state of this object from the given buffer. The buffer may contain more data after the
        structured element is parsed. This data is ignored.

        :param buffer: The buffer to read data from
        :param offset: The offset in the buffer where to start reading
        :param length: The amount of data we are allowed to read from the buffer
        :return: The number of bytes used from the buffer
        """
        my_offset, option_len = self.parse_option_header(buffer, offset, length)
        header_offset = my_offset

        # Parse the options
        self.options = []
        max_offset = option_len + header_offset  # The option_len field counts bytes *after* the header fields
        while max_offset > my_offset:
            used_buffer, option = NTPSubOption.parse(buffer, offset=offset + my_offset)
            self.options.append(option)
            my_offset += used_buffer

        if my_offset != max_offset:
            raise ValueError('Option length does not match the combined length of the parsed suboptions')

        self.validate()

        return my_offset

    def save(self) -> bytes:
        """
        Save the internal state of this object as a buffer.

        :return: The buffer with the data from this element
        """
        self.validate()

        options_buffer = bytearray()
        for option in self.options:
            options_buffer.extend(option.save())

        buffer = bytearray()
        buffer.extend(pack('!HH', self.option_type, len(options_buffer)))
        buffer.extend(options_buffer)
        return buffer


# Register the classes in this file
register(NTPServerAddressSubOption)
register(NTPMulticastAddressSubOption)
register(NTPServerFQDNSubOption)

# Specify which class may occur where
ClientServerMessage.add_may_contain(NTPServersOption)
NTPServersOption.add_may_contain(NTPSubOption, 1)



"""
Test the behaviour of listening sockets.
"""
import logging
from socket import AF_INET, AF_INET6, IPPROTO_UDP, IPPROTO_TCP
from ipaddress import IPv6Address
import unittest
from unittest.mock import Mock

from dhcpkit.ipv6 import SERVER_PORT, All_DHCP_Relay_Agents_and_Servers, CLIENT_PORT
from dhcpkit.ipv6.exceptions import ListeningSocketError, InvalidPacketError
from dhcpkit.ipv6.listening_socket import ListeningSocket
from dhcpkit.ipv6.messages import RelayForwardMessage, UnknownMessage, RelayReplyMessage, Message
from dhcpkit.ipv6.options import InterfaceIdOption, RelayMessageOption
from tests.ipv6.messages.test_advertise_message import advertise_message, advertise_packet
from tests.ipv6.messages.test_relay_forward_message import relayed_solicit_packet, relayed_solicit_message
from tests.ipv6.messages.test_relay_reply_message import relayed_advertise_message, relayed_advertise_packet
from tests.ipv6.messages.test_solicit_message import solicit_packet, solicit_message


class MockSocket:
    """
    Mock-up of a network socket for unit testing
    """

    def __init__(self, family: int, proto: int, address: str, port: int, interface_id: int, file_no: int):
        """
        Construct a new mock socket
        """
        self.family = family
        self.proto = proto
        self.getsockname = Mock(return_value=(address, port, 0, interface_id))
        self.file_no = file_no

        self.incoming_queue = []
        self.outgoing_queue = []

        self.pretend_sendto_fails = False

    def fileno(self) -> int:
        """
        Return a fake file descriptor
        """
        return self.file_no

    def add_to_incoming_queue(self, packet: bytes, sender: tuple):
        """
        Add this packet to the "received" queue

        :param packet: The bytes of the packet
        :param sender: The sender of the packet
        :type sender: (str, int, int, int)
        """
        self.incoming_queue.append((packet, sender))

    def recvfrom(self, bufsize: int) -> (bytes, (str, int, int, int)):
        """
        Pretend that this message was "received"

        :param bufsize: Length to truncate the packet to
        :return:
        """
        packet, sender = self.incoming_queue.pop()
        return packet[:bufsize], sender

    def sendto(self, packet: bytes, sender: tuple):
        """
        Pretend that this message was "sent"

        :param packet: The bytes of the packet
        :param sender: The sender of the packet
        :type sender: (str, int, int, int)
        """
        if self.pretend_sendto_fails:
            # Oops, we lost a byte
            packet = packet[:-1]

        self.outgoing_queue.append((packet, sender))
        return len(packet)

    def read_from_outgoing_queue(self) -> (bytes, (str, int, int, int)):
        """
        Get the first packet that was "sent"

        :return: The packet and the destination
        """
        return self.outgoing_queue.pop()


class ListeningSocketTestCase(unittest.TestCase):
    def test_constructor_link_local(self):
        link_local_socket = MockSocket(AF_INET6, IPPROTO_UDP, 'fe80::1%eth0', SERVER_PORT, 42, 1608)

        # noinspection PyTypeChecker
        listening_socket = ListeningSocket('eth0', link_local_socket, global_address=IPv6Address('2001:db8::1'))

        self.assertEqual(listening_socket.interface_name, 'eth0')
        self.assertEqual(listening_socket.interface_id, b'eth0')
        self.assertEqual(listening_socket.interface_index, 42)
        self.assertEqual(listening_socket.listen_socket, link_local_socket)
        self.assertEqual(listening_socket.listen_address, IPv6Address('fe80::1'))
        self.assertEqual(listening_socket.reply_socket, link_local_socket)
        self.assertEqual(listening_socket.reply_address, IPv6Address('fe80::1'))
        self.assertEqual(listening_socket.global_address, IPv6Address('2001:db8::1'))
        self.assertEqual(listening_socket.fileno(), 1608)

    def test_constructor_unicast(self):
        global_unicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, '2001:db8::1', SERVER_PORT, 42, 1608)

        # noinspection PyTypeChecker
        listening_socket = ListeningSocket('eth0', global_unicast_socket)

        self.assertEqual(listening_socket.interface_name, 'eth0')
        self.assertEqual(listening_socket.interface_id, b'eth0')
        self.assertEqual(listening_socket.interface_index, 42)
        self.assertEqual(listening_socket.listen_socket, global_unicast_socket)
        self.assertEqual(listening_socket.listen_address, IPv6Address('2001:db8::1'))
        self.assertEqual(listening_socket.reply_socket, global_unicast_socket)
        self.assertEqual(listening_socket.reply_address, IPv6Address('2001:db8::1'))
        self.assertEqual(listening_socket.global_address, IPv6Address('2001:db8::1'))
        self.assertEqual(listening_socket.fileno(), 1608)

    def test_constructor_multicast(self):
        multicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, All_DHCP_Relay_Agents_and_Servers, SERVER_PORT, 42, 1608)
        link_local_socket = MockSocket(AF_INET6, IPPROTO_UDP, 'fe80::1%eth0', SERVER_PORT, 42, 1608)

        # noinspection PyTypeChecker
        listening_socket = ListeningSocket('eth0', multicast_socket, link_local_socket,
                                           global_address=IPv6Address('2001:db8::1'))

        self.assertEqual(listening_socket.interface_name, 'eth0')
        self.assertEqual(listening_socket.interface_id, b'eth0')
        self.assertEqual(listening_socket.interface_index, 42)
        self.assertEqual(listening_socket.listen_socket, multicast_socket)
        self.assertEqual(listening_socket.listen_address, IPv6Address(All_DHCP_Relay_Agents_and_Servers))
        self.assertEqual(listening_socket.reply_socket, link_local_socket)
        self.assertEqual(listening_socket.reply_address, IPv6Address('fe80::1'))
        self.assertEqual(listening_socket.global_address, IPv6Address('2001:db8::1'))
        self.assertEqual(listening_socket.fileno(), 1608)

    def test_bad_proto(self):
        good_socket = MockSocket(AF_INET6, IPPROTO_UDP, '2001:db8::1', SERVER_PORT, 42, 1608)
        bad_socket_family = MockSocket(AF_INET, IPPROTO_UDP, '2001:db8::1', SERVER_PORT, 42, 1608)
        bad_socket_proto = MockSocket(AF_INET6, IPPROTO_TCP, '2001:db8::1', SERVER_PORT, 42, 1608)

        with self.assertRaisesRegex(ListeningSocketError, r'IPv6 UDP socket'):
            # noinspection PyTypeChecker
            ListeningSocket('eth0', bad_socket_family, good_socket)

        with self.assertRaisesRegex(ListeningSocketError, r'IPv6 UDP socket'):
            # noinspection PyTypeChecker
            ListeningSocket('eth0', bad_socket_proto, good_socket)

        with self.assertRaisesRegex(ListeningSocketError, r'IPv6 UDP socket'):
            # noinspection PyTypeChecker
            ListeningSocket('eth0', good_socket, bad_socket_family)

        with self.assertRaisesRegex(ListeningSocketError, r'IPv6 UDP socket'):
            # noinspection PyTypeChecker
            ListeningSocket('eth0', good_socket, bad_socket_proto)

    def test_bad_port(self):
        good_socket = MockSocket(AF_INET6, IPPROTO_UDP, '2001:db8::1', SERVER_PORT, 42, 1608)
        bad_socket = MockSocket(AF_INET6, IPPROTO_UDP, '2001:db8::1', CLIENT_PORT, 42, 1608)

        with self.assertRaisesRegex(ListeningSocketError, r'port 547'):
            # noinspection PyTypeChecker
            ListeningSocket('eth0', bad_socket, good_socket)

        with self.assertRaisesRegex(ListeningSocketError, r'port 547'):
            # noinspection PyTypeChecker
            ListeningSocket('eth0', good_socket, bad_socket)

    def test_global_address_finding(self):
        global_unicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, '2001:db8::2', SERVER_PORT, 42, 1608)
        link_local_socket = MockSocket(AF_INET6, IPPROTO_UDP, 'fe80::1%eth0', SERVER_PORT, 42, 1608)
        multicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, All_DHCP_Relay_Agents_and_Servers, SERVER_PORT, 42, 1608)

        # Test explicit global address
        # noinspection PyTypeChecker
        listening_socket = ListeningSocket('eth0', global_unicast_socket, global_address=IPv6Address('2001:db8::1'))
        self.assertEqual(listening_socket.global_address, IPv6Address('2001:db8::1'))

        # Test taking the listen socket address
        # noinspection PyTypeChecker
        listening_socket = ListeningSocket('eth0', global_unicast_socket)
        self.assertEqual(listening_socket.global_address, IPv6Address('2001:db8::2'))

        # Test no global address found
        with self.assertRaisesRegex(ListeningSocketError, r'Cannot determine global address'):
            # noinspection PyTypeChecker
            ListeningSocket('eth0', multicast_socket, link_local_socket)

    def test_interface_mismatch(self):
        multicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, All_DHCP_Relay_Agents_and_Servers, SERVER_PORT, 42, 1608)
        link_local_socket = MockSocket(AF_INET6, IPPROTO_UDP, 'fe80::1%eth0', SERVER_PORT, 41, 1608)

        with self.assertRaisesRegex(ListeningSocketError, r'same interface'):
            # noinspection PyTypeChecker
            ListeningSocket('eth0', multicast_socket, link_local_socket)

    def test_wildcard_address(self):
        wildcard_socket = MockSocket(AF_INET6, IPPROTO_UDP, '::', SERVER_PORT, 42, 1608)
        multicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, All_DHCP_Relay_Agents_and_Servers, SERVER_PORT, 42, 1608)

        with self.assertRaisesRegex(ListeningSocketError, r'wildcard'):
            # noinspection PyTypeChecker
            ListeningSocket('eth0', wildcard_socket)

        with self.assertRaisesRegex(ListeningSocketError, r'wildcard'):
            # noinspection PyTypeChecker
            ListeningSocket('eth0', multicast_socket, wildcard_socket, global_address=IPv6Address('2001:db8::1'))

    def test_global_reply_address(self):
        multicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, All_DHCP_Relay_Agents_and_Servers, SERVER_PORT, 42, 1608)
        global_unicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, '2001:db8::2', SERVER_PORT, 42, 1608)

        with self.assertRaisesRegex(ListeningSocketError, r'need link-local reply socket'):
            # noinspection PyTypeChecker
            ListeningSocket('eth0', multicast_socket, global_unicast_socket, global_address=IPv6Address('2001:db8::1'))

    def test_unwanted_reply_address(self):
        global_unicast_listen_socket = MockSocket(AF_INET6, IPPROTO_UDP, '2001:db8::1', SERVER_PORT, 42, 1608)
        global_unicast_reply_socket = MockSocket(AF_INET6, IPPROTO_UDP, '2001:db8::2', SERVER_PORT, 42, 1608)

        with self.assertRaisesRegex(ListeningSocketError, r"can't use separate reply socket"):
            # noinspection PyTypeChecker
            ListeningSocket('eth0', global_unicast_listen_socket, global_unicast_reply_socket)

    def test_receive_direct(self):
        multicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, All_DHCP_Relay_Agents_and_Servers, SERVER_PORT, 42, 1608)
        link_local_socket = MockSocket(AF_INET6, IPPROTO_UDP, 'fe80::1%eth0', SERVER_PORT, 42, 1608)

        # noinspection PyTypeChecker
        listening_socket = ListeningSocket('eth0', multicast_socket, link_local_socket,
                                           global_address=IPv6Address('2001:db8::1'))

        multicast_socket.add_to_incoming_queue(solicit_packet, ('2001:db8::babe', 546, 0, 42))
        with self.assertLogs(level=logging.DEBUG) as logged:
            received_message = listening_socket.recv_request()

        self.assertIsInstance(received_message, RelayForwardMessage)
        self.assertEqual(received_message.hop_count, 0)
        self.assertEqual(received_message.link_address, IPv6Address('2001:db8::1'))
        self.assertEqual(received_message.peer_address, IPv6Address('2001:db8::babe'))
        self.assertEqual(received_message.get_option_of_type(InterfaceIdOption).interface_id, b'eth0')
        self.assertEqual(received_message.relayed_message, solicit_message)

        log_output = '\n'.join(logged.output)
        self.assertRegex(log_output, r'Received SolicitMessage')
        self.assertRegex(log_output, r'from 2001:db8::babe')

    def test_receive_relayed(self):
        global_unicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, '2001:db8::1', SERVER_PORT, 42, 1608)

        # noinspection PyTypeChecker
        listening_socket = ListeningSocket('eth0', global_unicast_socket)

        global_unicast_socket.add_to_incoming_queue(relayed_solicit_packet, ('2001:db8::babe', 546, 0, 42))
        with self.assertLogs(level=logging.DEBUG) as logged:
            received_message = listening_socket.recv_request()

        self.assertIsInstance(received_message, RelayForwardMessage)
        self.assertEqual(received_message.hop_count, 2)
        self.assertEqual(received_message.link_address, IPv6Address('2001:db8::1'))
        self.assertEqual(received_message.peer_address, IPv6Address('2001:db8::babe'))
        self.assertEqual(received_message.get_option_of_type(InterfaceIdOption).interface_id, b'eth0')
        self.assertEqual(received_message.relayed_message, relayed_solicit_message)

        log_output = '\n'.join(logged.output)
        self.assertRegex(log_output, r'Received SolicitMessage')
        self.assertRegex(log_output, r'from fe80::3631:c4ff:fe3c:b2f1')
        self.assertRegex(log_output, r'via Fa2/3 of relay 2001:db8::babe')

    def test_receive_relayed_with_unprintable_interface_id(self):
        global_unicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, '2001:db8::1', SERVER_PORT, 42, 1608)

        # noinspection PyTypeChecker
        listening_socket = ListeningSocket('eth0', global_unicast_socket)

        # Start with a clean parse and then change interface-id
        new_message = Message.parse(relayed_solicit_packet)[1]
        new_message.inner_relay_message.get_option_of_type(InterfaceIdOption).interface_id = b'\x80\x81\x82'

        global_unicast_socket.add_to_incoming_queue(bytes(new_message.save()), ('2001:db8::babe', 546, 0, 42))
        with self.assertLogs(level=logging.DEBUG) as logged:
            listening_socket.recv_request()

        log_output = '\n'.join(logged.output)
        self.assertRegex(log_output, r'Received SolicitMessage')
        self.assertRegex(log_output, r'from fe80::3631:c4ff:fe3c:b2f1')
        self.assertRegex(log_output, r"via b'\\x80\\x81\\x82' of relay 2001:db8::babe")

    def test_receive_relayed_without_interface_id(self):
        global_unicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, '2001:db8::1', SERVER_PORT, 42, 1608)

        # noinspection PyTypeChecker
        listening_socket = ListeningSocket('eth0', global_unicast_socket)

        # Start with a clean parse and then remove interface-id
        new_message = Message.parse(relayed_solicit_packet)[1]
        interface_id_option = new_message.inner_relay_message.get_option_of_type(InterfaceIdOption)
        new_message.inner_relay_message.options.remove(interface_id_option)

        global_unicast_socket.add_to_incoming_queue(bytes(new_message.save()), ('2001:db8::babe', 546, 0, 42))
        with self.assertLogs(level=logging.DEBUG) as logged:
            listening_socket.recv_request()

        log_output = '\n'.join(logged.output)
        self.assertRegex(log_output, r'Received SolicitMessage')
        self.assertRegex(log_output, r'from fe80::3631:c4ff:fe3c:b2f1')
        self.assertRegex(log_output, r'via relay 2001:db8::babe')

    def test_receive_bad_message(self):
        multicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, All_DHCP_Relay_Agents_and_Servers, SERVER_PORT, 42, 1608)
        link_local_socket = MockSocket(AF_INET6, IPPROTO_UDP, 'fe80::1%eth0', SERVER_PORT, 42, 1608)

        # noinspection PyTypeChecker
        listening_socket = ListeningSocket('eth0', multicast_socket, link_local_socket,
                                           global_address=IPv6Address('2001:db8::1'))

        multicast_socket.add_to_incoming_queue(b'\x01ThisIsNotAValidDHCPv6Message', ('2001:db8::babe', 546, 0, 42))
        with self.assertRaisesRegex(InvalidPacketError, r"Invalid packet from \('2001:db8::babe', 546, 0, 42\)"):
            listening_socket.recv_request()

    def test_receive_unknown_message_type(self):
        multicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, All_DHCP_Relay_Agents_and_Servers, SERVER_PORT, 42, 1608)
        link_local_socket = MockSocket(AF_INET6, IPPROTO_UDP, 'fe80::1%eth0', SERVER_PORT, 42, 1608)

        # noinspection PyTypeChecker
        listening_socket = ListeningSocket('eth0', multicast_socket, link_local_socket,
                                           global_address=IPv6Address('2001:db8::1'))

        multicast_socket.add_to_incoming_queue(b'\xffThisIsNotAValidDHCPv6Message', ('2001:db8::babe', 546, 0, 42))
        received_message = listening_socket.recv_request()

        self.assertIsInstance(received_message, RelayForwardMessage)
        self.assertEqual(received_message.hop_count, 0)
        self.assertEqual(received_message.link_address, IPv6Address('2001:db8::1'))
        self.assertEqual(received_message.peer_address, IPv6Address('2001:db8::babe'))
        self.assertEqual(received_message.get_option_of_type(InterfaceIdOption).interface_id, b'eth0')
        self.assertIsInstance(received_message.relayed_message, UnknownMessage)
        self.assertEqual(received_message.relayed_message.message_type, 255)

    def test_send_direct(self):
        multicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, All_DHCP_Relay_Agents_and_Servers, SERVER_PORT, 42, 1608)
        link_local_socket = MockSocket(AF_INET6, IPPROTO_UDP, 'fe80::1%eth0', SERVER_PORT, 42, 1608)

        # noinspection PyTypeChecker
        listening_socket = ListeningSocket('eth0', multicast_socket, link_local_socket,
                                           global_address=IPv6Address('2001:db8::1'))

        outgoing_message = RelayReplyMessage(hop_count=0,
                                             link_address=IPv6Address('2001:db8::1'),
                                             peer_address=IPv6Address('fe80::babe'),
                                             options=[
                                                 InterfaceIdOption(interface_id=b'eth0'),
                                                 RelayMessageOption(relayed_message=advertise_message)
                                             ])

        with self.assertLogs(level=logging.DEBUG) as logged:
            success = listening_socket.send_reply(outgoing_message)
            self.assertTrue(success)

        # Nothing should be sent from a multicast socket
        with self.assertRaises(IndexError):
            multicast_socket.read_from_outgoing_queue()

        # It must be on the link local socket
        sent_packet, recipient = link_local_socket.read_from_outgoing_queue()
        self.assertEqual(sent_packet, advertise_packet)
        self.assertEqual(recipient, ('fe80::babe', CLIENT_PORT, 0, 42))

        log_output = '\n'.join(logged.output)
        self.assertRegex(log_output, r'Sent AdvertiseMessage')
        self.assertRegex(log_output, r'to fe80::babe')

    def test_failed_send_direct(self):
        multicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, All_DHCP_Relay_Agents_and_Servers, SERVER_PORT, 42, 1608)
        link_local_socket = MockSocket(AF_INET6, IPPROTO_UDP, 'fe80::1%eth0', SERVER_PORT, 42, 1608)
        link_local_socket.pretend_sendto_fails = True

        # noinspection PyTypeChecker
        listening_socket = ListeningSocket('eth0', multicast_socket, link_local_socket,
                                           global_address=IPv6Address('2001:db8::1'))

        outgoing_message = RelayReplyMessage(hop_count=0,
                                             link_address=IPv6Address('2001:db8::1'),
                                             peer_address=IPv6Address('fe80::babe'),
                                             options=[
                                                 InterfaceIdOption(interface_id=b'eth0'),
                                                 RelayMessageOption(relayed_message=advertise_message)
                                             ])

        with self.assertLogs(level=logging.DEBUG) as logged:
            success = listening_socket.send_reply(outgoing_message)
            self.assertFalse(success)

        # Nothing should be sent from a multicast socket
        with self.assertRaises(IndexError):
            multicast_socket.read_from_outgoing_queue()

        # It must be on the link local socket
        sent_packet, recipient = link_local_socket.read_from_outgoing_queue()
        self.assertNotEqual(sent_packet, advertise_packet)

        log_output = '\n'.join(logged.output)
        self.assertRegex(log_output, r'AdvertiseMessage')
        self.assertRegex(log_output, r'to fe80::babe')
        self.assertRegex(log_output, r'could not be sent')

    def test_send_relayed(self):
        multicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, All_DHCP_Relay_Agents_and_Servers, SERVER_PORT, 42, 1608)
        link_local_socket = MockSocket(AF_INET6, IPPROTO_UDP, 'fe80::1%eth0', SERVER_PORT, 42, 1608)

        # noinspection PyTypeChecker
        listening_socket = ListeningSocket('eth0', multicast_socket, link_local_socket,
                                           global_address=IPv6Address('2001:db8::1'))

        outgoing_message = RelayReplyMessage(hop_count=0,
                                             link_address=IPv6Address('2001:db8::1'),
                                             peer_address=IPv6Address('fe80::babe'),
                                             options=[
                                                 InterfaceIdOption(interface_id=b'eth0'),
                                                 RelayMessageOption(relayed_message=relayed_advertise_message)
                                             ])

        with self.assertLogs(level=logging.DEBUG) as logged:
            success = listening_socket.send_reply(outgoing_message)
            self.assertTrue(success)

        # Nothing should be sent from a multicast socket
        with self.assertRaises(IndexError):
            multicast_socket.read_from_outgoing_queue()

        # It must be on the link local socket
        sent_packet, recipient = link_local_socket.read_from_outgoing_queue()
        self.assertEqual(sent_packet, relayed_advertise_packet)
        self.assertEqual(recipient, ('fe80::babe', SERVER_PORT, 0, 42))

        log_output = '\n'.join(logged.output)
        self.assertRegex(log_output, r'Sent AdvertiseMessage')
        self.assertRegex(log_output, r'to fe80::3631:c4ff:fe3c:b2f1')
        self.assertRegex(log_output, r'via Fa2/3 of relay fe80::babe')

    def test_failed_send_relayed(self):
        multicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, All_DHCP_Relay_Agents_and_Servers, SERVER_PORT, 42, 1608)
        link_local_socket = MockSocket(AF_INET6, IPPROTO_UDP, 'fe80::1%eth0', SERVER_PORT, 42, 1608)
        link_local_socket.pretend_sendto_fails = True

        # noinspection PyTypeChecker
        listening_socket = ListeningSocket('eth0', multicast_socket, link_local_socket,
                                           global_address=IPv6Address('2001:db8::1'))

        outgoing_message = RelayReplyMessage(hop_count=0,
                                             link_address=IPv6Address('2001:db8::1'),
                                             peer_address=IPv6Address('fe80::babe'),
                                             options=[
                                                 InterfaceIdOption(interface_id=b'eth0'),
                                                 RelayMessageOption(relayed_message=relayed_advertise_message)
                                             ])

        with self.assertLogs(level=logging.DEBUG) as logged:
            success = listening_socket.send_reply(outgoing_message)
            self.assertFalse(success)

        # Nothing should be sent from a multicast socket
        with self.assertRaises(IndexError):
            multicast_socket.read_from_outgoing_queue()

        # It must be on the link local socket
        sent_packet, recipient = link_local_socket.read_from_outgoing_queue()
        self.assertNotEqual(sent_packet, relayed_advertise_packet)

        log_output = '\n'.join(logged.output)
        self.assertRegex(log_output, r'AdvertiseMessage')
        self.assertRegex(log_output, r'to fe80::3631:c4ff:fe3c:b2f1')
        self.assertRegex(log_output, r"via Fa2/3 of relay fe80::babe")

    def test_send_relayed_with_unprintable_interface_id(self):
        multicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, All_DHCP_Relay_Agents_and_Servers, SERVER_PORT, 42, 1608)
        link_local_socket = MockSocket(AF_INET6, IPPROTO_UDP, 'fe80::1%eth0', SERVER_PORT, 42, 1608)

        # noinspection PyTypeChecker
        listening_socket = ListeningSocket('eth0', multicast_socket, link_local_socket,
                                           global_address=IPv6Address('2001:db8::1'))

        # Start with a clean parse and then change interface-id
        new_message = Message.parse(relayed_advertise_packet)[1]
        new_message.inner_relay_message.get_option_of_type(InterfaceIdOption).interface_id = b'\x80\x81\x82'

        outgoing_message = RelayReplyMessage(hop_count=0,
                                             link_address=IPv6Address('2001:db8::1'),
                                             peer_address=IPv6Address('fe80::babe'),
                                             options=[
                                                 InterfaceIdOption(interface_id=b'eth0'),
                                                 RelayMessageOption(relayed_message=new_message)
                                             ])

        with self.assertLogs(level=logging.DEBUG) as logged:
            listening_socket.send_reply(outgoing_message)

        log_output = '\n'.join(logged.output)
        self.assertRegex(log_output, r'Sent AdvertiseMessage')
        self.assertRegex(log_output, r'to fe80::3631:c4ff:fe3c:b2f1')
        self.assertRegex(log_output, r"via b'\\x80\\x81\\x82' of relay fe80::babe")

    def test_send_relayed_without_interface_id(self):
        multicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, All_DHCP_Relay_Agents_and_Servers, SERVER_PORT, 42, 1608)
        link_local_socket = MockSocket(AF_INET6, IPPROTO_UDP, 'fe80::1%eth0', SERVER_PORT, 42, 1608)

        # noinspection PyTypeChecker
        listening_socket = ListeningSocket('eth0', multicast_socket, link_local_socket,
                                           global_address=IPv6Address('2001:db8::1'))

        # Start with a clean parse and then change interface-id
        new_message = Message.parse(relayed_advertise_packet)[1]
        interface_id_option = new_message.inner_relay_message.get_option_of_type(InterfaceIdOption)
        new_message.inner_relay_message.options.remove(interface_id_option)

        outgoing_message = RelayReplyMessage(hop_count=0,
                                             link_address=IPv6Address('2001:db8::1'),
                                             peer_address=IPv6Address('fe80::babe'),
                                             options=[
                                                 InterfaceIdOption(interface_id=b'eth0'),
                                                 RelayMessageOption(relayed_message=new_message)
                                             ])

        with self.assertLogs(level=logging.DEBUG) as logged:
            listening_socket.send_reply(outgoing_message)

        log_output = '\n'.join(logged.output)
        self.assertRegex(log_output, r'Sent AdvertiseMessage')
        self.assertRegex(log_output, r'to fe80::3631:c4ff:fe3c:b2f1')
        self.assertRegex(log_output, r"via relay fe80::babe")

    def test_send_unwrapped(self):
        multicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, All_DHCP_Relay_Agents_and_Servers, SERVER_PORT, 42, 1608)
        link_local_socket = MockSocket(AF_INET6, IPPROTO_UDP, 'fe80::1%eth0', SERVER_PORT, 42, 1608)

        # noinspection PyTypeChecker
        listening_socket = ListeningSocket('eth0', multicast_socket, link_local_socket,
                                           global_address=IPv6Address('2001:db8::1'))

        with self.assertRaisesRegex(ValueError, r'has to be wrapped'):
            listening_socket.send_reply(advertise_message)

    def test_send_badly_wrapped(self):
        multicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, All_DHCP_Relay_Agents_and_Servers, SERVER_PORT, 42, 1608)
        link_local_socket = MockSocket(AF_INET6, IPPROTO_UDP, 'fe80::1%eth0', SERVER_PORT, 42, 1608)

        # noinspection PyTypeChecker
        listening_socket = ListeningSocket('eth0', multicast_socket, link_local_socket,
                                           global_address=IPv6Address('2001:db8::1'))

        with self.assertRaisesRegex(ValueError, r'link-address does not match'):
            listening_socket.send_reply(relayed_advertise_message)

        # Fix the link-address so the test continues to the interface-id
        # noinspection PyTypeChecker
        listening_socket = ListeningSocket('eth0', multicast_socket, link_local_socket,
                                           global_address=IPv6Address('2001:db8:ffff:1::1'))

        with self.assertRaisesRegex(ValueError, r'interface-id in the reply does not match'):
            listening_socket.send_reply(relayed_advertise_message)

    def test_send_empty_wrapper(self):
        multicast_socket = MockSocket(AF_INET6, IPPROTO_UDP, All_DHCP_Relay_Agents_and_Servers, SERVER_PORT, 42, 1608)
        link_local_socket = MockSocket(AF_INET6, IPPROTO_UDP, 'fe80::1%eth0', SERVER_PORT, 42, 1608)

        # noinspection PyTypeChecker
        listening_socket = ListeningSocket('eth0', multicast_socket, link_local_socket,
                                           global_address=IPv6Address('2001:db8::1'))

        outgoing_message = RelayReplyMessage(hop_count=0,
                                             link_address=IPv6Address('2001:db8::1'),
                                             peer_address=IPv6Address('fe80::babe'),
                                             options=[
                                                 InterfaceIdOption(interface_id=b'eth0'),
                                             ])

        with self.assertRaisesRegex(ValueError, r'not contain a message'):
            listening_socket.send_reply(outgoing_message)


if __name__ == '__main__':
    unittest.main()



"""
Test the RelayForwardMessage implementation
"""
import unittest
import codecs
from ipaddress import IPv6Network, IPv6Address

from dhcpkit.ipv6.duids import LinkLayerDUID
from dhcpkit.ipv6.messages import SolicitMessage, RelayForwardMessage
from dhcpkit.ipv6.options import ElapsedTimeOption, ClientIdOption, RapidCommitOption, IANAOption, \
    ReconfigureAcceptOption, OptionRequestOption, OPTION_IA_NA, OPTION_VENDOR_OPTS, VendorClassOption, \
    RelayMessageOption, InterfaceIdOption
from dhcpkit.ipv6.extensions.prefix_delegation import IAPDOption, IAPrefixOption, OPTION_IA_PD
from dhcpkit.ipv6.extensions.dns import OPTION_DNS_SERVERS
from dhcpkit.ipv6.extensions.sntp import OPTION_SNTP_SERVERS
from dhcpkit.ipv6.extensions.remote_id import RemoteIdOption
from dhcpkit.ipv6.extensions.ntp import OPTION_NTP_SERVER
from dhcpkit.ipv6.extensions.sol_max_rt import OPTION_SOL_MAX_RT, OPTION_INF_MAX_RT
from tests.ipv6.messages import test_relay_server_message

""

# DHCPv6
#     Message type: Relay-forw (12)
#     Hop count: 1
#     Link address: 2001:db8:ffff:1::1 (2001:db8:ffff:1::1)
#     Peer address: fe80::3631:c4ff:fe3c:b2f1 (fe80::3631:c4ff:fe3c:b2f1)
#     Relay Message
#         Option: Relay Message (9)
#         Length: 194
#         Value: 0c0000000000000000000000000000000000fe8000000000...
#         DHCPv6
#             Message type: Relay-forw (12)
#             Hop count: 0
#             Link address: :: (::)
#             Peer address: fe80::3631:c4ff:fe3c:b2f1 (fe80::3631:c4ff:fe3c:b2f1)
#             Relay Message
#                 Option: Relay Message (9)
#                 Length: 121
#                 Value: 01f350d60008000200000001000a000300013431c43cb2f1...
#                 DHCPv6
#                     Message type: Solicit (1)
#                     Transaction ID: 0xf350d6
#                     Elapsed time
#                         Option: Elapsed time (8)
#                         Length: 2
#                         Value: 0000
#                         Elapsed time: 0 ms
#                     Client Identifier
#                         Option: Client Identifier (1)
#                         Length: 10
#                         Value: 000300013431c43cb2f1
#                         DUID: 000300013431c43cb2f1
#                         DUID Type: link-layer address (3)
#                         Hardware type: Ethernet (1)
#                         Link-layer address: 34:31:c4:3c:b2:f1
#                     Rapid Commit
#                         Option: Rapid Commit (14)
#                         Length: 0
#                     Identity Association for Non-temporary Address
#                         Option: Identity Association for Non-temporary Address (3)
#                         Length: 12
#                         Value: c43cb2f10000000000000000
#                         IAID: c43cb2f1
#                         T1: 0
#                         T2: 0
#                     Identity Association for Prefix Delegation
#                         Option: Identity Association for Prefix Delegation (25)
#                         Length: 41
#                         Value: c43cb2f10000000000000000001a00190000000000000000...
#                         IAID: c43cb2f1
#                         T1: 0
#                         T2: 0
#                         IA Prefix
#                             Option: IA Prefix (26)
#                             Length: 25
#                             Value: 000000000000000000000000000000000000000000000000...
#                             Preferred lifetime: 0
#                             Valid lifetime: 0
#                             Prefix length: 0
#                             Prefix address: :: (::)
#                     Reconfigure Accept
#                         Option: Reconfigure Accept (20)
#                         Length: 0
#                     Option Request
#                         Option: Option Request (6)
#                         Length: 16
#                         Value: 00170038001f00190003001100520053
#                         Requested Option code: DNS recursive name server (23)
#                         Requested Option code: NTP Server (56)
#                         Requested Option code: Simple Network Time Protocol Server (31)
#                         Requested Option code: Identity Association for Prefix Delegation (25)
#                         Requested Option code: Identity Association for Non-temporary Address (3)
#                         Requested Option code: Vendor-specific Information (17)
#                         Requested Option code: SOL_MAX_RT (82)
#                         Requested Option code: INF_MAX_RT (83)
#                     Vendor Class
#                         Option: Vendor Class (16)
#                         Length: 4
#                         Value: 00000368
#                         Enterprise ID: AVM GmbH (872)
#             Interface-Id
#                 Option: Interface-Id (18)
#                 Length: 5
#                 Value: 4661322f33
#                 Interface-ID: 4661322f33
#             Remote Identifier
#                 Option: Remote Identifier (37)
#                 Length: 22
#                 Value: 00000009020023000001000a0003000100211c7d486e
#                 Enterprise ID: ciscoSystems (9)
#                 Remote-ID: 020023000001000a0003000100211c7d486e
#     Interface-Id
#         Option: Interface-Id (18)
#         Length: 7
#         Value: 4769302f302f30
#         Interface-ID: 4769302f302f30
#     Remote Identifier
#         Option: Remote Identifier (37)
#         Length: 22
#         Value: 00000009020000000000000a0003000124e9b36e8100
#         Enterprise ID: ciscoSystems (9)
#         Remote-ID: 020000000000000a0003000124e9b36e8100

relayed_solicit_message = RelayForwardMessage(
    hop_count=1,
    link_address=IPv6Address('2001:db8:ffff:1::1'),
    peer_address=IPv6Address('fe80::3631:c4ff:fe3c:b2f1'),
    options=[
        RelayMessageOption(relayed_message=RelayForwardMessage(
            hop_count=0,
            link_address=IPv6Address('::'),
            peer_address=IPv6Address('fe80::3631:c4ff:fe3c:b2f1'),
            options=[
                RelayMessageOption(relayed_message=SolicitMessage(
                    transaction_id=bytes.fromhex('f350d6'),
                    options=[
                        ElapsedTimeOption(elapsed_time=0),
                        ClientIdOption(duid=LinkLayerDUID(hardware_type=1,
                                                          link_layer_address=bytes.fromhex('3431c43cb2f1'))),
                        RapidCommitOption(),
                        IANAOption(iaid=bytes.fromhex('c43cb2f1')),
                        IAPDOption(iaid=bytes.fromhex('c43cb2f1'), options=[
                            IAPrefixOption(prefix=IPv6Network('::/0')),
                        ]),
                        ReconfigureAcceptOption(),
                        OptionRequestOption(requested_options=[
                            OPTION_DNS_SERVERS,
                            OPTION_NTP_SERVER,
                            OPTION_SNTP_SERVERS,
                            OPTION_IA_PD,
                            OPTION_IA_NA,
                            OPTION_VENDOR_OPTS,
                            OPTION_SOL_MAX_RT,
                            OPTION_INF_MAX_RT,
                        ]),
                        VendorClassOption(enterprise_number=872),
                    ],
                )),
                InterfaceIdOption(interface_id=b'Fa2/3'),
                RemoteIdOption(enterprise_number=9, remote_id=bytes.fromhex('020023000001000a0003000100211c7d486e')),
            ])
        ),
        InterfaceIdOption(interface_id=b'Gi0/0/0'),
        RemoteIdOption(enterprise_number=9, remote_id=bytes.fromhex('020000000000000a0003000124e9b36e8100')),
    ],
)

relayed_solicit_packet = codecs.decode('0c0120010db8ffff0001000000000000'
                                       '0001fe800000000000003631c4fffe3c'
                                       'b2f1000900c20c000000000000000000'
                                       '0000000000000000fe80000000000000'
                                       '3631c4fffe3cb2f10009007901f350d6'
                                       '0008000200000001000a000300013431'
                                       'c43cb2f1000e00000003000cc43cb2f1'
                                       '000000000000000000190029c43cb2f1'
                                       '0000000000000000001a001900000000'
                                       '00000000000000000000000000000000'
                                       '00000000000014000000060010001700'
                                       '38001f00190003001100520053001000'
                                       '0400000368001200054661322f330025'
                                       '001600000009020023000001000a0003'
                                       '000100211c7d486e001200074769302f'
                                       '302f3000250016000000090200000000'
                                       '00000a0003000124e9b36e8100', 'hex')


class RelayedSolicitMessageTestCase(test_relay_server_message.RelayServerMessageTestCase):
    def setUp(self):
        self.packet_fixture = relayed_solicit_packet
        self.message_fixture = relayed_solicit_message
        self.parse_packet()


if __name__ == '__main__':
    unittest.main()



"""
Base class for pkg_resources based registries
"""
import collections
import logging

import pkg_resources

logger = logging.getLogger(__name__)


class Registry(collections.UserDict):
    """
    Base class for registries
    """

    entry_point = 'dhcpkit.NONE'
    """The name of the entry_point group"""

    def __init__(self):
        """
        A custom dictionary that initialises itself with the entry points from pkg_resources
        """
        super().__init__()

        # Try all the entry points
        entry_points = pkg_resources.iter_entry_points(group=self.entry_point)
        for entry_point in entry_points:
            # If the name is a string with an integer then convert it to a real integer
            try:
                name = int(entry_point.name)
            except ValueError:
                name = entry_point.name

            if name in self.data:
                logger.warning("Multiple entry points found for {} {}, using {}".format(
                    self.__class__.__name__, name, self.data[name]))
                continue

            try:
                # Load the entry point and store it
                self.data[name] = entry_point.load()
            except ImportError:
                # Ok, this one isn't working, skip it
                logger.error("Entry points {} for {} could not be loaded".format(
                    entry_point, self.__class__.__name__))
                continue



"""
An object to hold everything related to a request/response transaction
"""
import logging
from ipaddress import IPv6Address

from dhcpkit.ipv6.extensions.prefix_delegation import IAPDOption
from dhcpkit.ipv6.messages import Message, RelayForwardMessage, ClientServerMessage, UnknownMessage, RelayReplyMessage
from dhcpkit.ipv6.options import Option, IANAOption, IATAOption

logger = logging.getLogger(__name__)


class TransactionBundle:
    """
    A bundle with all data about a transaction. This makes it much easier to pass around multiple pieces of information.

    :type incoming_message: Message
    :type received_over_multicast: bool
    :type request: ClientServerMessage
    :type incoming_relay_messages: list[RelayForwardMessage]
    :type response: ClientServerMessage
    :type outgoing_relay_messages: list[RelayReplyMessage]
    :type handled_options: list[Option]
    """

    def __init__(self, incoming_message: Message, received_over_multicast: bool, allow_rapid_commit: bool = False):
        self.allow_rapid_commit = allow_rapid_commit
        """Allow rapid commit? May be set to True on creation, may be set to False by option handlers, not vice versa"""

        self.incoming_message = incoming_message
        """The incoming message including the relay chain"""

        self.received_over_multicast = received_over_multicast
        """A flag indicating whether the client used multicast to contact the server"""

        # Convenience properties for easy access to the request and chain without having to walk the chain every time
        self.request = None
        """The incoming request without the relay messages"""

        self.incoming_relay_messages = []
        """The chain of relay messages starting with the one closest to the client"""

        self.request, self.incoming_relay_messages = self.split_relay_chain(incoming_message)

        self.response = None
        """This is where the user puts the response :class:`.ClientServerMessage`"""

        self.outgoing_relay_messages = None
        """This is where the user puts the reply relay chain by calling :meth:`create_outgoing_relay_messages`"""

        self.handled_options = []
        """A list of options from the request that have been handled, only applies to IA type options"""

    def mark_handled(self, option: Option):
        """
        Mark the given option as handled
        :param option:
        :return:
        """
        if option not in self.handled_options:
            self.handled_options.append(option)

    def get_unanswered_ia_options(self) -> [IANAOption or IATAOption]:
        """
        Get a list of all IANAOptions and IATAOptions in the request that have no counterpart in the response

        :return: The list of unanswered IANAOptions and IATAOptions
        :rtype: list[IANAOption or IATAOption]
        """
        # Make a list of requested IANAOptions
        return [option for option in self.request.options
                if isinstance(option, (IANAOption, IATAOption)) and option not in self.handled_options]

    def get_unanswered_iana_options(self) -> [IANAOption]:
        """
        Get a list of all IANAOptions in the request that have no counterpart in the response

        :return: The list of unanswered IANAOptions
        :rtype: list[IANAOption]
        """
        # Make a list of requested IANAOptions
        return [option for option in self.request.options
                if isinstance(option, IANAOption) and option not in self.handled_options]

    def get_unanswered_iata_options(self) -> [IATAOption]:
        """
        Get a list of all IATAOptions in the request that have no counterpart in the response

        :return: The list of unanswered IATAOptions
        :rtype: list[IATAOption]
        """
        # Make a list of requested IANAOptions
        return [option for option in self.request.options
                if isinstance(option, IATAOption) and option not in self.handled_options]

    def get_unanswered_iapd_options(self) -> [IAPDOption]:
        """
        Get a list of all IANAOptions in the request that have no counterpart in the response

        :return: The list of unanswered IAPDOptions
        :rtype: list[IAPDOption]
        """
        # Make a list of requested IANAOptions
        return [option for option in self.request.options
                if isinstance(option, IAPDOption) and option not in self.handled_options]

    @staticmethod
    def split_relay_chain(message: Message) -> (ClientServerMessage, [RelayForwardMessage]):
        """
        Separate the relay chain from the actual request message.

        :param message: The incoming message
        :returns: The request and the chain of relay messages starting with the one closest to the client
        """
        relay_messages = []
        while isinstance(message, RelayForwardMessage):
            relay_messages.insert(0, message)
            message = message.relayed_message

        # Check if we could actually read the message
        if isinstance(message, UnknownMessage):
            logger.warning("Received an unrecognised message of type {}".format(message.message_type))
            return None, None

        # Check that this message is a client->server message
        if not isinstance(message, ClientServerMessage) or not message.from_client_to_server:
            logger.warning("A server should not receive {} from a client".format(message.__class__.__name__))
            return None, None

        # Save it as the request
        return message, relay_messages

    def get_link_address(self) -> IPv6Address:
        """
        Find the link address that identifies where this request is coming from
        """
        # Start with the relay closest to the client and keep looking until a useful address is found
        for relay in self.incoming_relay_messages:
            # Some relays (i.e. LDRA: :rfc:`6221`) don't have a useful link-address
            if not relay.link_address.is_unspecified and \
                    not relay.link_address.is_loopback and \
                    not relay.link_address.is_link_local:
                # This looks useful
                return relay.link_address

        # Nothing useful...
        return IPv6Address('::')

    def create_outgoing_relay_messages(self):
        """
        Create a plain chain of RelayReplyMessages for the current response
        """
        if not self.response:
            raise ValueError("Cannot create outgoing relay messages without a response")

        outgoing_message = self.incoming_relay_messages[-1].wrap_response(self.response)
        self.outgoing_relay_messages = []
        while isinstance(outgoing_message, RelayReplyMessage):
            self.outgoing_relay_messages.insert(0, outgoing_message)
            outgoing_message = outgoing_message.relayed_message

    @property
    def outgoing_message(self):
        """
        Wrap the response in a relay chain if necessary
        """
        if self.response is None:
            # No response is ok
            return None

        if not self.response.from_server_to_client:
            logger.error("A server should not send {} to a client".format(self.response.__class__.__name__))
            return None

        if self.incoming_relay_messages and not self.outgoing_relay_messages:
            # No outgoing relay messages, but we had incoming relay messages: auto-create a plain relay chain
            self.create_outgoing_relay_messages()

        if self.outgoing_relay_messages:
            # Make sure the right response is in the relay messages (in case someone overwrites :attr:`response` without
            # updating the contents of the relay messages as well
            self.outgoing_relay_messages[0].relayed_message = self.response

            # Send the relay messages
            return self.outgoing_relay_messages[-1]
        else:
            # Send the plain response
            return self.response



"""
Option handlers for the DNS options defined in dhcpkit.ipv6.extensions.sol_max_rt
"""

import configparser

from dhcpkit.ipv6.extensions.sol_max_rt import SolMaxRTOption, InfMaxRTOption
from dhcpkit.ipv6.option_handlers import OverwritingOptionHandler, OptionHandler


class SolMaxRTOptionHandler(OverwritingOptionHandler):
    """
    Handler for putting SolMaxRTOption in responses
    """

    def __init__(self, sol_max_rt: int):
        option = SolMaxRTOption(sol_max_rt=sol_max_rt)
        option.validate()

        super().__init__(option)

    @classmethod
    def from_config(cls, section: configparser.SectionProxy, option_handler_id: str = None) -> OptionHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :param section: The configuration section
        :param option_handler_id: Optional extra identifier
        :return: A handler object
        :rtype: OptionHandler
        """
        sol_max_rt = section.getint('sol-max-rt')
        if sol_max_rt is None:
            raise configparser.NoOptionError('sol-max-rt', section.name)

        return cls(sol_max_rt)


class InfMaxRTOptionHandler(OverwritingOptionHandler):
    """
    Handler for putting InfMaxRTOption in responses
    """

    def __init__(self, inf_max_rt: int):
        option = InfMaxRTOption(inf_max_rt=inf_max_rt)
        option.validate()

        super().__init__(option)

    @classmethod
    def from_config(cls, section: configparser.SectionProxy, option_handler_id: str = None) -> OptionHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :param section: The configuration section
        :param option_handler_id: Optional extra identifier
        :return: A handler object
        :rtype: OptionHandler
        """
        inf_max_rt = section.getint('inf-max-rt')
        if inf_max_rt is None:
            raise configparser.NoOptionError('inf-max-rt', section.name)

        return cls(inf_max_rt)
